'''
Helper functions for the library
'''

from __future__ import annotations

import logging
import struct
import subprocess
import time
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .exceptions import (  # noqa:F401
    EDIDParseError,
    MaxRetriesExceededError,
    ScreenBrightnessError,
    format_exc,
)
from .types import DisplayIdentifier, Generator, IntPercentage, Percentage

_logger = logging.getLogger(__name__)

MONITOR_MANUFACTURER_CODES = {
    'AAC': 'AcerView',
    'ACI': 'Asus (ASUSTeK Computer Inc.)',
    'ACR': 'Acer',
    'ACT': 'Targa',
    'ADI': 'ADI Corporation',
    'AIC': 'AG Neovo',
    'ALX': 'Anrecson',
    'AMW': 'AMW',
    'AOC': 'AOC',
    'API': 'Acer America Corp.',
    'APP': 'Apple Computer',
    'ART': 'ArtMedia',
    'AST': 'AST Research',
    'AUO': 'Asus',
    'BMM': 'BMM',
    'BNQ': 'BenQ',
    'BOE': 'BOE Display Technology',
    'CMO': 'Acer',
    'CPL': 'Compal',
    'CPQ': 'Compaq',
    'CPT': 'Chunghwa Picture Tubes, Ltd.',
    'CTX': 'CTX',
    'DEC': 'DEC',
    'DEL': 'Dell',
    'DPC': 'Delta',
    'DWE': 'Daewoo',
    'ECS': 'ELITEGROUP Computer Systems',
    'EIZ': 'EIZO',
    'ELS': 'ELSA',
    'ENC': 'EIZO',
    'EPI': 'Envision',
    'FCM': 'Funai',
    'FUJ': 'Fujitsu',
    'FUS': 'Fujitsu-Siemens',
    'GSM': 'LG Electronics',
    'GWY': 'Gateway 2000',
    'GBT': 'Gigabyte',
    'HEI': 'Hyundai',
    'HIQ': 'Hyundai ImageQuest',
    'HIT': 'Hyundai',
    'HPN': 'HP',
    'HSD': 'Hannspree Inc',
    'HSL': 'Hansol',
    'HTC': 'Hitachi/Nissei',
    'HWP': 'HP',
    'IBM': 'IBM',
    'ICL': 'Fujitsu ICL',
    'IFS': 'InFocus',
    'IQT': 'Hyundai',
    'IVM': 'Iiyama',
    'KDS': 'Korea Data Systems',
    'KFC': 'KFC Computek',
    'LEN': 'Lenovo',
    'LGD': 'Asus',
    'LKM': 'ADLAS / AZALEA',
    'LNK': 'LINK Technologies, Inc.',
    'LPL': 'Fujitsu',
    'LTN': 'Lite-On',
    'MAG': 'MAG InnoVision',
    'MAX': 'Belinea',
    'MEI': 'Panasonic',
    'MEL': 'Mitsubishi Electronics',
    'MIR': 'miro Computer Products AG',
    'MSI': 'MSI',
    'MS_': 'Panasonic',
    'MTC': 'MITAC',
    'NAN': 'Nanao',
    'NEC': 'NEC',
    'NOK': 'Nokia Data',
    'NVD': 'Fujitsu',
    'OPT': 'Optoma',
    'OQI': 'OPTIQUEST',
    'PBN': 'Packard Bell',
    'PCK': 'Daewoo',
    'PDC': 'Polaroid',
    'PGS': 'Princeton Graphic Systems',
    'PHL': 'Philips',
    'PRT': 'Princeton',
    'REL': 'Relisys',
    'SAM': 'Samsung',
    'SAN': 'Samsung',
    'SBI': 'Smarttech',
    'SEC': 'Hewlett-Packard',
    'SGI': 'SGI',
    'SMC': 'Samtron',
    'SMI': 'Smile',
    'SNI': 'Siemens Nixdorf',
    'SNY': 'Sony',
    'SPT': 'Sceptre',
    'SRC': 'Shamrock',
    'STN': 'Samtron',
    'STP': 'Sceptre',
    'SUN': 'Sun Microsystems',
    'TAT': 'Tatung',
    'TOS': 'Toshiba',
    'TRL': 'Royal Information Company',
    'TSB': 'Toshiba',
    'UNK': 'Unknown',
    'UNM': 'Unisys Corporation',
    'VSC': 'ViewSonic',
    'WTC': 'Wen Technology',
    'ZCM': 'Zenith',
    '_YV': 'Fujitsu',
}


class BrightnessMethod(ABC):
    @classmethod
    @abstractmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        '''
        Return information about detected displays.

        Args:
            display (.types.DisplayIdentifier): the specific display to return
                information about. This parameter is passed to `filter_monitors`

        Returns:
            A list of dictionaries, each representing a detected display.
            Each returned dictionary will have the following keys:
            - name (`str`): the name of the display
            - model (`str`): the model of the display
            - manufacturer (`str`): the name of the display manufacturer
            - manufacturer_id (`str`): the three letter manufacturer code (see `MONITOR_MANUFACTURER_CODES`)
            - serial (`str`): the serial of the display OR some other unique identifier
            - edid (`str`): the EDID string for the display
            - method (`BrightnessMethod`): the brightness method associated with this display
            - index (`int`): the index of the display, relative to the brightness method
        '''
        ...

    @classmethod
    @abstractmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        '''
        Args:
            display: the index of the specific display to query.
                If unspecified, all detected displays are queried

        Returns:
            A list of `.types.IntPercentage` values, one for each
            queried display
        '''
        ...

    @classmethod
    @abstractmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        '''
        Args:
            value (.types.IntPercentage): the new brightness value
            display: the index of the specific display to adjust.
                If unspecified, all detected displays are adjusted
        '''
        ...


class BrightnessMethodAdv(BrightnessMethod):
    @classmethod
    @abstractmethod
    def _gdi(cls) -> List[dict]:
        '''
        Similar to `BrightnessMethod.get_display_info` except this method will also
        return unsupported displays, indicated by an `unsupported: bool` property
        in the returned dict
        '''
        ...


class __Cache:
    '''class to cache data with a short shelf life'''

    def __init__(self):
        self.logger = _logger.getChild(f'{self.__class__.__name__}_{id(self)}')
        self.enabled = True
        self._store: Dict[str, Tuple[Any, float]] = {}

    def expire(self, key: Optional[str] = None, startswith: Optional[str] = None):
        '''
        @private

        Runs through all keys in the cache and removes any expired items.
        Can optionally specify additional keys that should be removed.

        Args:
            key: a specific key to remove. `KeyError` exceptions are suppressed if this key doesn't exist.
            startswith: remove any keys that start with this string
        '''
        if key is not None:
            try:
                del self._store[key]
                self.logger.debug(f'delete key {key!r}')
            except KeyError:
                pass

        for k, v in tuple(self._store.items()):
            if startswith is not None and k.startswith(startswith):
                del self._store[k]
                self.logger.debug(f'delete keys {startswith=}')
                continue
            if v[1] < time.time():
                del self._store[k]
                self.logger.debug(f'delete expired key {k}')

    def get(self, key: str) -> Any:
        if not self.enabled:
            return None
        self.expire()
        if key not in self._store:
            self.logger.debug(f'{key!r} not present in cache')
            return None
        return self._store[key][0]

    def store(self, key: str, value: Any, expires: float = 1):
        if not self.enabled:
            return
        self.logger.debug(f'cache set {key!r}, {expires=}')
        self._store[key] = (value, expires + time.time())


class EDID:
    '''
    Simple structure and method to extract display serial and name from an EDID string.
    '''

    EDID_FORMAT: str = (
        '>'  # big-endian
        '8s'  # constant header (8 bytes)
        'H'  # manufacturer id (2 bytes)
        'H'  # product id (2 bytes)
        'I'  # serial number (4 bytes)
        'B'  # manufactoring week (1 byte)
        'B'  # manufactoring year (1 byte)
        'B'  # edid version (1 byte)
        'B'  # edid revision (1 byte)
        'B'  # video input type (1 byte)
        'B'  # horizontal size in cm (1 byte)
        'B'  # vertical size in cm (1 byte)
        'B'  # display gamma (1 byte)
        'B'  # supported features (1 byte)
        '10s'  # colour characteristics (10 bytes)
        'H'  # supported timings (2 bytes)
        'B'  # reserved timing (1 byte)
        '16s'  # EDID supported timings (16 bytes)
        '18s'  # timing / display descriptor block 1 (18 bytes)
        '18s'  # timing / display descriptor block 2 (18 bytes)
        '18s'  # timing / display descriptor block 3 (18 bytes)
        '18s'  # timing / display descriptor block 4 (18 bytes)
        'B'  # extension flag (1 byte)
        'B'  # checksum (1 byte)
    )
    '''
    The byte structure for EDID strings, taken from
    [pyedid](https://github.com/jojonas/pyedid/blob/2382910d968b2fa8de1fab495fbbdfebcdb39f19/pyedid/edid.py#L21),
    [Copyright 2019-2020 Jonas Lieb, Davydov Denis](https://github.com/jojonas/pyedid/blob/master/LICENSE).
    '''
    SERIAL_DESCRIPTOR = bytes.fromhex('00 00 00 ff 00')
    NAME_DESCRIPTOR = bytes.fromhex('00 00 00 fc 00')

    @classmethod
    def parse(cls, edid: Union[bytes, str]) -> Tuple[Union[str, None], ...]:
        '''
        Takes an EDID string and parses some relevant information from it according to the
        [EDID 1.4](https://en.wikipedia.org/wiki/Extended_Display_Identification_Data#EDID_1.4_data_format)
        specification on Wikipedia.

        Args:
            edid (bytes or str): the EDID, can either be raw bytes or
                a hex formatted string (00 ff ff ff ff...)

        Returns:
            tuple[str | None]: A tuple of 5 items representing the display's manufacturer ID,
                manufacturer, model, name, serial in that order.
                If any of these values are unable to be determined, they will be None.
                Otherwise, expect a string

        Raises:
            EDIDParseError: if the EDID info cannot be unpacked
            TypeError: if `edid` is not `str` or `bytes`

        Example:
            ```python
            import screen_brightness_control as sbc

            edid = sbc.list_monitors_info()[0]['edid']
            manufacturer_id, manufacturer, model, name, serial = sbc.EDID.parse(edid)

            print('Manufacturer:', manufacturer_id or 'Unknown')
            print('Model:', model or 'Unknown')
            print('Name:', name or 'Unknown')
            ```
        '''
        # see https://en.wikipedia.org/wiki/Extended_Display_Identification_Data#EDID_1.4_data_format
        if isinstance(edid, str):
            edid = bytes.fromhex(edid)
        elif not isinstance(edid, bytes):
            raise TypeError(f'edid must be of type bytes or str, not {type(edid)!r}')

        try:
            blocks = struct.unpack(cls.EDID_FORMAT, edid)
        except struct.error as e:
            raise EDIDParseError('cannot unpack edid') from e

        # split mfg_id (2 bytes) into 3 letters, 5 bits each (ignoring reserved bit)
        mfg_id_chars = (
            blocks[1] >> 10,  # First 6 bits (reserved bit at start is always 0)
            (blocks[1] >> 5) & 0b11111,  # isolate next 5 bits from first 11 using bitwise AND
            blocks[1] & 0b11111,  # Last five bits
        )
        # turn numbers into ascii
        mfg_id = ''.join(chr(i + 64) for i in mfg_id_chars)

        # now grab the manufacturer name
        mfg_lookup = _monitor_brand_lookup(mfg_id)
        if mfg_lookup is not None:
            manufacturer = mfg_lookup[1]
        else:
            manufacturer = None

        serial = None
        name = None
        for descriptor_block in blocks[17:21]:
            # decode the serial
            if descriptor_block.startswith(cls.SERIAL_DESCRIPTOR):
                # strip descriptor bytes and trailing whitespace
                serial_bytes = descriptor_block[len(cls.SERIAL_DESCRIPTOR) :].rstrip()
                serial = serial_bytes.decode()

            # decode the monitor name
            elif descriptor_block.startswith(cls.NAME_DESCRIPTOR):
                # strip descriptor bytes and trailing whitespace
                name_bytes = descriptor_block[len(cls.NAME_DESCRIPTOR) :].rstrip()
                name = name_bytes.decode()

        # now try to figure out what model the display is
        model = None
        if name is not None:
            if manufacturer is not None and name.startswith(manufacturer):
                # eg: 'BenQ GL2450H' -> 'GL2450H'
                model = name.replace(manufacturer, '', 1).strip()

            # if previous method did not work (or if we don't know the manufacturer),
            # try taking last word of name
            if not model:
                try:
                    # eg: 'BenQ GL2450H' -> ['BenQ', 'GL2450H']
                    model = name.strip().rsplit(' ', 1)[1]
                except IndexError:
                    # If the name does not include model information then
                    # give it something generic
                    model = 'Generic Monitor'

        return mfg_id, manufacturer, model, name, serial

    @staticmethod
    def hexdump(file: str) -> str:
        '''
        Returns a hexadecimal string of binary data from a file

        Args:
            file (str): the file to read

        Returns:
            str: one long hex string

        Example:
            ```python
            from screen_brightness_control import EDID

            print(EDID.hexdump('/sys/class/backlight/intel_backlight/device/edid'))
            # '00ffffffffffff00...'
            ```
        '''
        with open(file, 'rb') as f:
            hex_str = ''.join(f'{char:02x}' for char in f.read())

        return hex_str


def check_output(command: List[str], max_tries: int = 1) -> bytes:
    '''
    Run a command with retry management built in.

    Args:
        command: the command to run
        max_tries: the maximum number of retries to allow before raising an error

    Returns:
        The output from the command
    '''
    tries = 1
    while True:
        try:
            output = subprocess.check_output(command, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as e:
            if tries >= max_tries:
                raise MaxRetriesExceededError(f'process failed after {tries} tries', e) from e
            tries += 1
            time.sleep(0.04 if tries < 5 else 0.5)
        else:
            if tries > 1:
                _logger.debug(f'command {command} took {tries}/{max_tries} tries')
            return output


def logarithmic_range(start: int, stop: int, step: int = 1) -> Generator[int, None, None]:
    '''
    A `range`-like function that yields a sequence of integers following
    a logarithmic curve (`y = 10 ^ (x / 50)`) from `start` (inclusive) to
    `stop` (inclusive).

    This is useful because it skips many of the higher percentages in the
    sequence where single percent brightness changes are hard to notice.

    This function is designed to deal with brightness percentages, and so
    will never return a value less than 0 or greater than 100.

    Args:
        start: the start of your percentage range
        stop: the end of your percentage range
        step: the increment per iteration through the sequence

    Yields:
        int
    '''
    start = int(max(0, start))
    stop = int(min(100, stop))

    if start == stop or abs(stop - start) <= 1:
        yield start
    else:
        value_range = stop - start

        def direction(x):
            return x if step > 0 else 100 - x

        last_yielded = None
        x: float
        for x in range(start, stop + 1, step):
            # get difference from base point
            x -= start
            # calculate progress through our range as a percentage
            x = (x / value_range) * 100
            # convert along logarithmic curve (inverse of y = 50log(x)) to another percentage
            x = 10 ** (direction(x) / 50)
            # apply this percentage to our range and add back starting offset
            x = int(((direction(x) / 100) * value_range) + start)

            if x == last_yielded:
                continue
            yield x
            last_yielded = x


@lru_cache(maxsize=None)
def _monitor_brand_lookup(search: str) -> Union[Tuple[str, str], None]:
    '''internal function to search the monitor manufacturer codes dict'''
    keys = tuple(MONITOR_MANUFACTURER_CODES.keys())
    keys_lower = tuple(map(str.lower, keys))
    values = tuple(MONITOR_MANUFACTURER_CODES.values())
    search = search.lower()

    if search in keys_lower:
        index = keys_lower.index(search)
    else:
        values_lower = tuple(map(str.lower, values))
        if search in values_lower:
            index = values_lower.index(search)
        else:
            return None
    return keys[index], values[index]


def percentage(
    value: Percentage, current: Optional[Union[int, Callable[[], int]]] = None, lower_bound: int = 0
) -> IntPercentage:
    '''
    Convenience function to convert a brightness value into a percentage. Can handle
    integers, floats and strings. Also can handle relative strings (eg: `'+10'` or `'-10'`)

    Args:
        value: the brightness value to convert
        current: the current brightness value or a function that returns the current brightness
            value. Used when dealing with relative brightness values
        lower_bound: the minimum value the brightness can be set to

    Returns:
        `.types.IntPercentage`: The new brightness percentage, between `lower_bound` and 100
    '''
    if isinstance(value, str) and ('+' in value or '-' in value):
        if callable(current):
            current = current()
        value = int(float(value)) + int(float(str(current)))
    else:
        value = int(float(str(value)))

    return min(100, max(lower_bound, value))



'''
Submodule containing types and type aliases used throughout the library.

Splitting these definitions into a seperate submodule allows for detailed
explanations and verbose type definitions, without cluttering up the rest
of the library.

This file is also useful for wrangling types based on the current Python
version.
'''

import sys
from typing import Union

# a bunch of typing classes were deprecated in Python 3.9
# in favour of collections.abc (https://www.python.org/dev/peps/pep-0585/)
if sys.version_info[1] >= 9:
    from collections.abc import Generator
else:
    from typing import Generator  # noqa: F401

IntPercentage = int
'''
An integer between 0 and 100 (inclusive) that represents a brightness level.
Other than the implied bounds, this is just a normal integer.
'''
Percentage = Union[IntPercentage, str]
'''
An `IntPercentage` or a string representing an `IntPercentage`.

String values may come in two forms:
- Absolute values: for example `'40'` converts directly to `int('40')`
- Relative values: strings prefixed with `+`/`-` will be interpreted relative to the
    current brightness level. In this case, the integer value of your string will be added to the
    current brightness level.
    For example, if the current brightness is 50%, a value of `'+40'` would imply 90% brightness
    and a value of `'-40'` would imply 10% brightness.

Relative brightness values will usually be resolved by the `.helpers.percentage` function.
'''


DisplayIdentifier = Union[int, str]
'''
Something that can be used to identify a particular display.
Can be any one of the following properties of a display:
- uid (str)
- edid (str)
- serial (str)
- name (str)
- index (int)

See `.Display` for descriptions of each property and its type
'''



import itertools
import subprocess
import time
from unittest.mock import Mock, call, mock_open

import pytest
from pytest_mock import MockerFixture

import screen_brightness_control as sbc
from screen_brightness_control.helpers import EDID, _monitor_brand_lookup, percentage

from .helpers import fake_edid


class TestCache:
    @pytest.fixture(scope='function')
    def cache(self):
        # have to use getattr otherwise python mangles the name due to lead dunder
        return getattr(sbc.helpers, '__Cache')()

    def test_get(self, cache):
        c_time = time.time()
        cache._store.update({'a': (123, c_time + 1), 'b': (456, c_time - 1)})

        assert cache.get('a') == 123
        assert 'a' in cache._store
        assert cache.get('b') is None
        # key should have been deleted as expired
        assert 'b' not in cache._store

    @pytest.mark.parametrize('expires', [1, 3, 5, -1])
    def test_store(self, cache, expires: int):
        c_time = time.time()
        cache.store('abc', 123, expires=expires)
        assert 'abc' in cache._store
        item = cache._store['abc']
        assert item[0] == 123
        assert (c_time + expires) - item[1] < 0.1

    def test_expire(self, cache):
        cache._store.update({'a': (123, 0), 'b': (123, 0), 'bc': (123, 0), 'def': (123, 0)})
        cache.expire('a')
        assert 'a' not in cache._store
        cache.expire(startswith='b')
        assert 'b' not in cache._store
        assert 'bc' not in cache._store
        # `expire` now expires all out of date keys automatically
        assert 'def' not in cache._store


class TestEDID:
    class TestParse:
        @pytest.fixture(
            params=[
                ('DEL', 'Dell U2211H', 'DELL12345'),
                ('BNQ', 'BenQ GL2450H', 'benqserialnum'),
                ('MSI', 'MSI G32CQ4', 'abc123MSI'),
            ]
        )
        def edid_with_params(self, request: pytest.FixtureRequest):
            return fake_edid(*request.param), *request.param

        @pytest.fixture
        def edid(self, edid_with_params):
            return edid_with_params[0]

        def test_returns_tuple(self, edid_with_params: tuple):
            '''
            Test the return type, return values and their meaning
            '''
            edid, mfg_id_in, name_in, serial_in = edid_with_params

            result = EDID.parse(edid)
            assert isinstance(result, tuple) and len(result) == 5

            mfg_id, manufacturer, model, name, serial = result
            assert mfg_id == mfg_id_in
            assert manufacturer == sbc.helpers.MONITOR_MANUFACTURER_CODES[mfg_id_in], (
                'manufacturer should be looked up correctly'
            )
            assert f'{manufacturer} {model}' == name == name_in, 'manufacturer and model should match the name'
            assert serial == serial_in

        def test_accepts_str_and_bytes(self, edid: str):
            EDID.parse(edid)
            EDID.parse(bytes.fromhex(edid))

            with pytest.raises(TypeError):
                EDID.parse(12345)  # type: ignore

        def test_invalid_edid_struct_raises_error(self):
            with pytest.raises(sbc.helpers.EDIDParseError):
                EDID.parse('00ff000000')

        class TestMfgId:
            @pytest.mark.parametrize('mfg_id_in', sbc.helpers.MONITOR_MANUFACTURER_CODES.keys())
            def test_monitor_manufacturer_id_is_parsed(self, mfg_id_in: str):
                '''
                Test that this can be parsed on its own, without any other info attached
                '''
                mfg_id, manufacturer, *_ = EDID.parse(fake_edid(mfg_id_in, 'a', 'b'))
                assert mfg_id == mfg_id_in
                assert manufacturer == sbc.helpers.MONITOR_MANUFACTURER_CODES[mfg_id_in]

            def test_invalid_input_returns_none(self):
                mfg_id, manufacturer, *_ = EDID.parse(fake_edid('ABC', 'a', 'b'))

                assert mfg_id == 'ABC'
                assert manufacturer is None

        @pytest.mark.parametrize('descriptor_block', ['name', 'serial'])
        def test_descriptor_blocks_are_optional(self, descriptor_block):
            info = {'name': 'Dell Sample1', 'serial': 'ABCD1234'}
            del info[descriptor_block]
            *_, name, serial = EDID.parse(fake_edid('DEL', **info))
            assert name == info.get('name', None)
            assert serial == info.get('serial', None)

        class TestModel:
            def test_model_is_none_when_name_missing(self):
                *_, model, name, _ = EDID.parse(fake_edid('DEL', serial='abc123'))
                assert name is None
                assert model is None

            def test_model_derived_from_name(self):
                model = EDID.parse(fake_edid('DEL', 'Dell Model1'))[2]
                assert model == 'Model1'

            def test_model_derived_when_unknown_manufacturer(self):
                model = EDID.parse(fake_edid('ABC', 'Dell Model1'))[2]
                assert model == 'Model1'

                # highly unlikely scenario given name has max len of 13 chars including
                # the manufacturer name
                model = EDID.parse(fake_edid('ABC', 'Nokia Data M1'))[2]
                assert model == 'M1', 'model should be taken from last word in display name'

            def test_generic_model_name_assigned_when_name_malformed(self):
                '''
                Sometimes a name can be malformed and only contain the manufacturer's name.

                See [this comment](https://github.com/Crozzers/screen_brightness_control/issues/19#issuecomment-1228525188)
                on issue #19 on the repo.
                '''
                model = EDID.parse(fake_edid('ABC', 'Dell '))[2]
                assert model == 'Generic Monitor'

    def test_hex_dump(self, mocker: MockerFixture):
        edid = fake_edid('DEL', 'Dell U2211H', 'ABCD1234')
        edid_bytes = bytes.fromhex(edid)
        mock_file = mocker.patch.object(sbc.helpers, 'open', mocker.mock_open(read_data=edid_bytes), spec=True)
        result = EDID.hexdump('edid_file')
        mock_file.assert_called_once_with('edid_file', 'rb')
        assert result == edid


class TestCheckOutput:
    def test_runs_command(self, mocker: MockerFixture):
        mock = mocker.patch.object(subprocess, 'check_output', Mock(return_value=b'test output'))
        command = ['do', 'nothing']
        assert sbc.helpers.check_output(command) == b'test output'
        assert mock.mock_calls[0].args[0] == command

    def test_retries(self, mocker: MockerFixture):
        command = ['do', 'nothing']

        mock = mocker.patch.object(
            subprocess, 'check_output', Mock(side_effect=subprocess.CalledProcessError(1, command))
        )

        with pytest.raises(sbc.exceptions.MaxRetriesExceededError):
            sbc.helpers.check_output(command, max_tries=3)

        assert len(mock.mock_calls) == 3 and all(call.args[0] == command for call in mock.mock_calls), (
            'command should have been tried 3 times'
        )


class TestLogarithmicRange:
    @pytest.fixture(
        params=[(0, 100), (0, 10), (29, 77), (99, 100), (0, 50), (50, 100), (0, 25), (25, 50), (50, 75), (75, 100)]
    )
    def bounds(self, request: pytest.FixtureRequest):
        return request.param

    @pytest.fixture
    def log_and_base_range(self, bounds):
        return list(sbc.logarithmic_range(*bounds)), list(range(*bounds))

    def test_returns_less_values_than_builtin_range(self, log_and_base_range):
        '''
        Main point of log range was to skip higher values, so it should always return less
        or the same number of values as `builtins.range`
        '''
        log_range, base_range = log_and_base_range

        assert len(log_range) <= len(base_range)

    def test_range_is_within_set_bounds(self, bounds, log_and_base_range):
        log_range = log_and_base_range[0]
        lower_bound, upper_bound = bounds

        # to GE chech here rather than EQ because input of 0 will return 1
        # as the min value, due to how the maths works out :/
        assert min(log_range) >= lower_bound
        # like with normal range, it won't reach the upper bound
        assert max(log_range) <= upper_bound

    def test_always_yields_int(self, log_and_base_range):
        log_range = log_and_base_range[0]

        assert all(isinstance(i, int) for i in log_range)

    @pytest.mark.parametrize('step', [1, 5, 10, -1, -5, -10])
    def test_step_kwarg(self, step):
        log_range = list(sbc.helpers.logarithmic_range(0, 100, step=step))

        diffs = [log_range[i + 1] - log_range[i] for i in range(len(log_range) - 2)]

        assert all(i != 0 for i in diffs), 'log range should never return the same number twice in a row'
        # positive * positive = positive and negative * negative = positive. Diff should always have
        # same sign as step, so diff * step should always be positive
        assert all(i * step > 0 for i in diffs), 'diff should match sign of step'

        if abs(step) != 1:
            base_range = list(sbc.logarithmic_range(0, 100, step=1))
            assert len(log_range) < len(base_range), 'bigger steps should yield less numbers'

    def test_skip_intervals(self, bounds):
        l_bound, u_bound = bounds

        if u_bound - l_bound < 2:
            pytest.skip('can only test skip interval when bounds are >2 apart')

        log_range = list(sbc.helpers.logarithmic_range(l_bound, u_bound))

        # lower value items have a lower diff than higher items
        assert log_range[1] - log_range[0] <= log_range[-1] - log_range[-2]

        log_range = list(sbc.helpers.logarithmic_range(u_bound, l_bound, -1))

        # higher value items have higher diff than lower items
        assert log_range[0] - log_range[1] >= log_range[-2] - log_range[-1]


class TestMonitorBrandLookup:
    def get_all_ids(self, manufacturer: str):
        # sometimes, multiple ids correspond to the same name (eg: Fujitsu)
        return [k for k, v in sbc.helpers.MONITOR_MANUFACTURER_CODES.items() if v.lower() == manufacturer.lower()]

    def test_returns_tuple_of_mfg_id_and_name(self):
        assert _monitor_brand_lookup('DEL') == _monitor_brand_lookup('Dell') == ('DEL', 'Dell')

    def test_bidirectional_lookups(self, subtests):
        for mfg_id, manufacturer in sbc.helpers.MONITOR_MANUFACTURER_CODES.items():
            with subtests.test(mfg_id=mfg_id, manufacturer=manufacturer):
                all_ids = self.get_all_ids(manufacturer)

                id_lookup = _monitor_brand_lookup(mfg_id)
                assert id_lookup is not None, 'ID lookup should be successful'
                assert id_lookup[0] in all_ids and id_lookup[1] == manufacturer, (
                    'ID lookup should return valid ID and manufacturer name'
                )

                name_lookup = _monitor_brand_lookup(manufacturer)
                assert name_lookup is not None, 'name lookup should be successful'
                assert name_lookup[1] == manufacturer and name_lookup[0] in all_ids, (
                    'name lookup should return valid ID and manufacturer name'
                )

    def test_lookup_names_are_case_corrected(self, subtests):
        for manufacturer in sbc.helpers.MONITOR_MANUFACTURER_CODES.values():
            for variation in [
                manufacturer,
                manufacturer.upper(),
                manufacturer.lower(),
                manufacturer.lower().capitalize(),
            ]:
                with subtests.test(manufacturer=manufacturer, variation=variation):
                    all_ids = self.get_all_ids(variation)
                    lookup = _monitor_brand_lookup(variation)
                    assert lookup is not None and lookup[0] in all_ids and lookup[1].lower() == variation.lower()

    def test_invalid_lookups(self):
        assert _monitor_brand_lookup('NUL') is None
        assert _monitor_brand_lookup('NotReal') is None


class TestPercentage:
    def test_numeric_values(self):
        # int
        assert percentage(100) == 100
        assert percentage(0) == 0
        assert percentage(50) == 50
        assert percentage(99) == 99

        # float
        assert percentage(100.0) == 100  # type: ignore
        assert percentage(99.99999999) == 99  # type: ignore
        assert percentage(1.5) == 1  # type: ignore

        # str - numeric-ish
        assert percentage('10') == 10
        assert percentage('55.555') == 55
        assert percentage('12.125') == 12

    def test_relative_values(self):
        assert percentage('+10', current=10) == 20
        assert percentage('-5', current=30) == 25
        assert percentage('-21', current=lambda: 99) == 78
        assert percentage('+50', current=lambda: 50) == 100
        assert percentage('-10.5', current=100, lower_bound=10) == 90

    def test_bounds(self):
        assert percentage(101) == 100
        assert percentage(1000) == 100
        assert percentage(-1) == 0
        assert percentage(-19999) == 0
        assert percentage('-100', current=0) == 0
        assert percentage('+1000000', current=0) == 100

        assert percentage(0, lower_bound=1) == 1
        assert percentage('-10', current=10, lower_bound=1) == 1

    def test_invalid_types(self):
        with pytest.raises(ValueError):
            percentage([123])  # type: ignore
        with pytest.raises(ValueError):
            percentage('123{')  # type: ignore



import glob
import os
import re
from typing import Type
from unittest.mock import Mock, call

import pytest
from pytest import MonkeyPatch
from pytest_mock import MockerFixture

import screen_brightness_control as sbc
from screen_brightness_control import linux
from screen_brightness_control.helpers import BrightnessMethod

from .helpers import BrightnessMethodTest
from .mocks.linux_mock import MockI2C, mock_check_output


class TestSysFiles(BrightnessMethodTest):
    @pytest.fixture
    def patch_get_display_info(self, mocker: MockerFixture):
        '''Mock everything needed to get `SysFiles.get_display_info` to run'''

        def listdir(dir: str):
            if 'subsystem' in dir:
                return ['intel_backlight', 'acpi_video0']
            return ['edp1']

        def isfile(file: str):
            return not file.endswith('device/edid')

        mocker.patch.object(os, 'listdir', Mock(side_effect=listdir), spec=True)
        mocker.patch.object(os.path, 'isdir', Mock(return_value=True), spec=True)
        mocker.patch.object(os.path, 'isfile', Mock(side_effect=isfile), spec=True)
        mocker.patch.object(sbc.linux, 'open', mocker.mock_open(read_data='100'), spec=True)

    @pytest.fixture
    def patch_get_brightness(self, mocker: MockerFixture, patch_get_display_info):
        pass

    @pytest.fixture
    def patch_set_brightness(self, mocker: MockerFixture, patch_get_display_info):
        pass

    @pytest.fixture
    def method(self) -> Type[BrightnessMethod]:
        return sbc.linux.SysFiles

    class TestGetDisplayInfo(BrightnessMethodTest.TestGetDisplayInfo):
        def test_returned_dicts_contain_required_keys(self, method):
            super().test_returned_dicts_contain_required_keys(method, extras={'scale': float, 'path': str})

        def test_display_filtering(self, mocker: MockerFixture, original_os_module, method):
            return super().test_display_filtering(mocker, original_os_module, method, extras={'include': ['path']})

        @pytest.mark.parametrize('max_brightness', (100, 200, 50))
        def test_brightness_scale(self, mocker: MockerFixture, method: Type[BrightnessMethod], max_brightness):
            mocker.patch.object(sbc.linux, 'open', mocker.mock_open(read_data=str(max_brightness)), spec=True)
            display = method.get_display_info()[0]
            assert display['scale'] == max_brightness / 100

        def test_empty_edid(self, mocker: MockerFixture, method: Type[BrightnessMethod]):
            mocker.patch.object(os.path, 'isfile', Mock(return_value=True), spec=True)
            mocker.patch.object(sbc.linux.EDID, 'hexdump', Mock(return_value=''), spec=True)

            displays = method.get_display_info()

            assert displays[0]['name'] == 'intel_backlight' and displays[0]['edid'] is None

    class TestGetBrightness(BrightnessMethodTest.TestGetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestGetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests):
                mock = mocker.patch.object(sbc.linux, 'open', mocker.mock_open(read_data='100'), spec=True)
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.get_brightness(display=index)
                        mock.assert_called_once_with(os.path.join(display['path'], 'brightness'), 'r')
                        mock.reset_mock()

            def test_without(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info):
                mock = mocker.patch.object(sbc.linux, 'open', mocker.mock_open(read_data='100'), spec=True)
                method.get_brightness()
                paths = [os.path.join(device['path'], 'brightness') for device in freeze_display_info]
                called_paths = [i[0][0] for i in mock.call_args_list]
                assert paths == called_paths

        @pytest.mark.parametrize('brightness', (100, 0, 50, 99))
        @pytest.mark.parametrize('scale', (1, 2, 0.5, 8))
        def test_brightness_is_scaled(
            self, mocker: MockerFixture, method: Type[BrightnessMethod], brightness: int, scale: float
        ):
            display = method.get_display_info()[0]
            display['scale'] = scale
            mocker.patch.object(method, 'get_display_info', Mock(return_value=[display]), spec=True)
            mocker.patch.object(sbc.linux, 'open', mocker.mock_open(read_data=str(brightness)), spec=True)

            assert method.get_brightness()[0] == brightness // scale

    class TestSetBrightness(BrightnessMethodTest.TestSetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestSetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests):
                mock = mocker.patch.object(sbc.linux, 'open', mocker.mock_open(), spec=True)
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.set_brightness(100, display=index)
                        mock.assert_called_once_with(os.path.join(display['path'], 'brightness'), 'w')
                        mock().write.assert_called_once_with('100')
                        mock.reset_mock()

            def test_without(
                self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests
            ):
                mock = mocker.patch.object(sbc.linux, 'open', mocker.mock_open(), spec=True)
                method.set_brightness(100)
                write = mock().write

                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        mock.assert_any_call(os.path.join(display['path'], 'brightness'), 'w')
                        assert write.call_args_list[index][0][0] == '100'


class TestI2C(BrightnessMethodTest):
    @pytest.fixture(scope='function', autouse=True)
    def cleanup(self, method: linux.I2C):
        method._max_brightness_cache = {}

    @pytest.fixture
    def patch_get_display_info(self, mocker: MockerFixture):
        def path_exists(path: str):
            return re.match(r'/dev/i2c-\d+', path) is not None

        mocker.patch.object(glob, 'glob', Mock(return_value=['/dev/i2c-0', '/dev/i2c-1']), spec=True)
        mocker.patch.object(os.path, 'exists', Mock(side_effect=path_exists), spec=True)
        mocker.patch.object(linux.I2C, 'I2CDevice', MockI2C.MockI2CDevice, spec=True)

    @pytest.fixture
    def patch_get_brightness(self, mocker: MockerFixture, patch_get_display_info):
        mocker.patch.object(linux.I2C, 'DDCInterface', MockI2C.MockDDCInterface, spec=True)

    @pytest.fixture
    def patch_set_brightness(self, mocker: MockerFixture, patch_get_display_info, patch_get_brightness):
        pass

    @pytest.fixture
    def method(self):
        return linux.I2C

    class TestGetDisplayInfo(BrightnessMethodTest.TestGetDisplayInfo):
        def test_returned_dicts_contain_required_keys(self, method: Type[BrightnessMethod]):
            return super().test_returned_dicts_contain_required_keys(method, {'i2c_bus': str})

        def test_display_filtering(self, mocker: MockerFixture, original_os_module, method):
            return super().test_display_filtering(mocker, original_os_module, method, {'include': ['i2c_bus']})

    class TestGetBrightness(BrightnessMethodTest.TestGetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestGetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests):
                spy = mocker.spy(method, 'DDCInterface')
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.get_brightness(display=index)
                        spy.assert_called_once_with(display['i2c_bus'])
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info):
                spy = mocker.spy(method, 'DDCInterface')
                method.get_brightness()
                paths = [device['i2c_bus'] for device in freeze_display_info]
                called_devices = [i[0][0] for i in spy.call_args_list]
                assert paths == called_devices

    class TestSetBrightness(BrightnessMethodTest.TestSetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestSetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests):
                spy = mocker.spy(method, 'DDCInterface')
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.set_brightness(100, display=index)
                        # one call for populating max brightness cache, another for setting brightness
                        spy.assert_has_calls([call(display['i2c_bus'])] * 2)
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info):
                spy = mocker.spy(method, 'DDCInterface')
                method.set_brightness(100)
                paths = [device['i2c_bus'] for device in freeze_display_info]
                called_devices = [i[0][0] for i in spy.call_args_list]
                # one call for populating max brightness cache, another for setting brightness, for each display
                assert sorted(called_devices) == sorted(paths * 2)


class TestXRandr(BrightnessMethodTest):
    @pytest.fixture
    def patch_get_display_info(self, mocker: MockerFixture, monkeypatch: MonkeyPatch):
        mock = Mock(side_effect=mock_check_output, spec=True)
        mocker.patch.object(sbc.helpers, 'check_output', mock)
        mocker.patch.object(sbc.linux, 'check_output', mock)
        # remove wayland from env to bypass compat checks
        monkeypatch.delitem(os.environ, 'WAYLAND_DISPLAY', raising=False)

    @pytest.fixture
    def patch_get_brightness(self, patch_get_display_info):
        pass

    @pytest.fixture
    def patch_set_brightness(self, patch_get_display_info):
        pass

    @pytest.fixture
    def method(self):
        return linux.XRandr

    class TestGetDisplayInfo(BrightnessMethodTest.TestGetDisplayInfo):
        def test_display_filtering(self, mocker: MockerFixture, original_os_module, method):
            super().test_display_filtering(mocker, original_os_module, method, extras={'include': ['interface']})

        def test_brightness_kwarg(self, method: Type[linux.XRandr]):
            assert all('brightness' not in display for display in method.get_display_info())
            with_brightness = method.get_display_info(brightness=True)
            assert all('brightness' in display for display in with_brightness)
            assert all(isinstance(d['brightness'], int) and 0 <= d['brightness'] <= 100 for d in with_brightness)

        def test_wayland(self, method: Type[linux.XRandr], monkeypatch: MonkeyPatch):
            monkeypatch.setitem(os.environ, 'WAYLAND_DISPLAY', 'wayland-0')
            assert method.get_display_info() == [], 'wayland displays not supported by xrandr'

    class TestGetBrightness(BrightnessMethodTest.TestGetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestGetBrightness.TestDisplayKwarg):
            # xrandr polls all displays at once in `get_display_info`, so `test_with*`
            # is not necessary. Providing stub implementations here to avoid NotImplementedError
            def test_with(self):
                pass

            def test_without(self):
                pass

    class TestSetBrightness(BrightnessMethodTest.TestSetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestSetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[linux.XRandr], freeze_display_info, subtests):
                spy = mocker.spy(sbc.linux, 'check_output')
                for index, monitor in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.set_brightness(100, display=index)
                        command = spy.call_args_list[0][0][0]
                        assert command.index('--output') == command.index(monitor['interface']) - 1
                        assert command.index('--brightness') == command.index('1.0') - 1
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method: Type[linux.XRandr], freeze_display_info):
                spy = mocker.spy(sbc.linux, 'check_output')
                method.set_brightness(100)
                interfaces = [i['interface'] for i in freeze_display_info]
                called_interfaces = [
                    cmd[cmd.index('--output') + 1] for cmd in map(lambda x: x[0][0], spy.call_args_list)
                ]
                assert sorted(interfaces) == sorted(called_interfaces)


class TestDDCUtil(BrightnessMethodTest):
    @pytest.fixture
    def patch_get_display_info(self, mocker: MockerFixture):
        mock = Mock(side_effect=mock_check_output, spec=True)
        mocker.patch.object(sbc.helpers, 'check_output', mock)
        mocker.patch.object(sbc.linux, 'check_output', mock)

    @pytest.fixture
    def patch_get_brightness(self, patch_get_display_info):
        pass

    @pytest.fixture
    def patch_set_brightness(self, patch_get_display_info):
        pass

    @pytest.fixture
    def method(self):
        return linux.DDCUtil

    class TestGetDisplayInfo(BrightnessMethodTest.TestGetDisplayInfo):
        def test_display_filtering(self, mocker: MockerFixture, original_os_module, method):
            return super().test_display_filtering(mocker, original_os_module, method, extras={'include': ['i2c_bus']})

    class TestGetBrightness(BrightnessMethodTest.TestGetBrightness):
        # TODO: tests for brightness scaling
        @pytest.fixture(autouse=True, scope='function')
        def patch(self, patch_get_brightness):
            sbc.linux.__cache__._store = {}

        class TestDisplayKwarg(BrightnessMethodTest.TestGetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests):
                spy = mocker.spy(sbc.linux, 'check_output')
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.get_brightness(display=index)
                        spy.assert_called_once()
                        command = spy.call_args_list[0][0][0]
                        assert command.index('-b') == command.index(str(display['bus_number'])) - 1
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info):
                spy = mocker.spy(sbc.linux, 'check_output')
                method.get_brightness()
                buses = [str(d['bus_number']) for d in freeze_display_info]
                called_buses = [i[i.index('-b') + 1] for i in map(lambda x: x[0][0], spy.call_args_list)]
                assert buses == called_buses

    class TestSetBrightness(BrightnessMethodTest.TestSetBrightness):
        @pytest.fixture(autouse=True, scope='function')
        def patch(self, patch_set_brightness):
            sbc.linux.__cache__._store = {}

        class TestDisplayKwarg(BrightnessMethodTest.TestSetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info, subtests):
                spy = mocker.spy(sbc.linux, 'check_output')
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.set_brightness(100, display=index)
                        spy.assert_called_once()
                        command = spy.call_args_list[0][0][0]
                        assert command.index('-b') == command.index(str(display['bus_number'])) - 1
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method: Type[BrightnessMethod], freeze_display_info):
                spy = mocker.spy(sbc.linux, 'check_output')
                method.set_brightness(100)
                buses = [str(d['bus_number']) for d in freeze_display_info]
                called_buses = [i[i.index('-b') + 1] for i in map(lambda x: x[0][0], spy.call_args_list)]
                assert buses == called_buses



'''
A small helper module to assist with debugging the screen_brightness_control library
'''

import logging
import platform
import traceback


def info() -> dict:
    '''
    Gather and return information that may (or may not) be useful for debugging
    issues with the library.

    All of the information returned by this function has, at some point, been
    requested and used as part of the debugging process of various
    [issues](https://github.com/Crozzers/screen_brightness_control/issues).
    '''
    import screen_brightness_control as sbc

    # configure logging
    logger = logging.getLogger(__name__).getChild('info')

    debug_info = {'version': sbc.__version__, 'platform': platform.system(), 'file': sbc.__file__}

    logger.debug('gathering list of all monitors')

    try:
        all_monitors = sbc.list_monitors_info(allow_duplicates=True, unsupported=True)
    except Exception:
        all_monitors = traceback.format_exc()
    finally:
        debug_info['all_monitors'] = [{'info': i} for i in all_monitors]

    logger.debug('filtering the list of monitors')

    try:
        if isinstance(all_monitors, list):
            all_monitors = [i for i in all_monitors if not i.get('unsupported')]
            monitor_info = sbc.filter_monitors(haystack=all_monitors)
        else:
            monitor_info = []
    except Exception:
        monitor_info = traceback.format_exc()
    finally:
        debug_info['filtered_monitors'] = monitor_info

    if not isinstance(all_monitors, str):  # if it is not a formatted traceback error msg
        for monitor_entry in debug_info['all_monitors']:
            monitor = monitor_entry['info']

            if not isinstance(monitor_info, str):  # if it is not a formatted traceback error msg
                if monitor in monitor_info:
                    monitor_entry['global_index'] = monitor_info.index(monitor)
                else:
                    monitor_entry['global_index'] = None

            logger.debug(f'get_brightness for monitor {monitor["method"].__name__}:{monitor["index"]}')

            try:
                brightness = monitor['method'].get_brightness(display=monitor['index'])
            except Exception:
                brightness = traceback.format_exc()
            finally:
                monitor_entry['get_brightness'] = brightness

            logger.debug(f'set_brightness for monitor {monitor["method"].__name__}:{monitor["index"]}')

            try:
                output = monitor['method'].set_brightness(
                    brightness[0] if isinstance(brightness, list) else 100, display=monitor['index']
                )
            except Exception:
                output = traceback.format_exc()
            finally:
                monitor_entry['set_brightness'] = output

    debug_info['methods'] = []
    for name, method in sbc.get_methods().items():
        logger.debug(f'getting display info for method: {name}')
        current = {'name': name, 'class': repr(method)}

        try:
            displays = method.get_display_info()
        except Exception:
            displays = traceback.format_exc()
        else:
            current['displays'] = []
            for d in displays:
                current_display = {}
                if not isinstance(all_monitors, str) and d in all_monitors:
                    current_display['index'] = all_monitors.index(d)
                if not isinstance(monitor_info, str) and d in monitor_info:
                    current_display['global_index'] = monitor_info.index(d)

                if not current_display:
                    current_display = d

                current['displays'].append(current_display)

        debug_info['methods'].append(current)

    if platform.system() == 'Windows':  # windows specific debug info
        logger.debug('windows specific debug info')
        debug_info['windows'] = {'wmi_monitor_id_output': sbc._OS_MODULE.wmi.WMI(namespace='wmi').WmiMonitorID()}
        try:
            enum_displays = [
                sbc._OS_MODULE.win32api.EnumDisplayDevices(
                    sbc._OS_MODULE.win32api.GetMonitorInfo(i[0])['Device'], 0, 1
                ).DeviceID
                for i in sbc._OS_MODULE.win32api.EnumDisplayMonitors()
            ]
        except Exception:
            enum_displays = traceback.format_exc()
        finally:
            debug_info['windows']['win32api_enum_display_device_ids'] = enum_displays

        try:
            physical_monitors = list(map(repr, sbc._OS_MODULE.VCP.iter_physical_monitors()))
        except Exception:
            physical_monitors = traceback.format_exc()
        finally:
            debug_info['windows']['physical_monitor_handles'] = physical_monitors

    return debug_info



import dataclasses
import threading
import time
from copy import deepcopy
from timeit import timeit
from typing import Any, Dict, List, cast
from unittest.mock import Mock, call

import pytest
from pytest_mock import MockerFixture

import screen_brightness_control as sbc

from .helpers import BrightnessFunctionTest
from .mocks import os_module_mock


class TestGetBrightness(BrightnessFunctionTest):
    @pytest.fixture
    def operation_type(self):
        return 'get'

    def test_returns_list_of_int(self):
        '''Check return types and integer bounds'''
        brightness = sbc.get_brightness()
        assert isinstance(brightness, list)
        assert all(isinstance(i, int) for i in brightness)
        assert all(0 <= i <= 100 for i in brightness)  # type: ignore


class TestSetBrightness(BrightnessFunctionTest):
    @pytest.fixture
    def operation_type(self):
        return 'set'

    def test_returns_none_by_default(self):
        assert sbc.set_brightness(100) is None

    def test_no_return_kwarg(self):
        result = sbc.set_brightness(100, no_return=False)
        assert result is not None
        assert isinstance(result, list) and all(i is None or (isinstance(i, int) and 0 <= i <= 100) for i in result), (
            'result should be a list of int|None and any ints should be between 0 and 100'
        )

    @pytest.mark.parametrize('os_name', ['Windows', 'Linux'])
    class TestLowerBound:
        percentage_spy: Mock
        brightness_spy: Mock
        lower_bound: int

        @pytest.fixture(autouse=True, scope='function')
        def patch(self, mocker: MockerFixture, os_name: str):
            mocker.patch.object(sbc.platform, 'system', new=lambda: os_name)
            self.percentage_spy = mocker.spy(sbc, 'percentage')
            self.brightness_spy = mocker.spy(sbc, '__brightness')
            self.lower_bound = 1 if os_name == 'Linux' else 0

        def test_lower_bound_applied(self):
            sbc.set_brightness(0)
            self.percentage_spy.assert_called_once_with(0, lower_bound=self.lower_bound)

        def test_force_kwarg(self):
            sbc.set_brightness(0)
            self.percentage_spy.assert_called_once_with(0, lower_bound=self.lower_bound)
            self.percentage_spy.reset_mock()

            sbc.set_brightness(0, force=True)
            self.percentage_spy.assert_called_once_with(0, lower_bound=0)

    class TestRelativeValues:
        setter_spy: Mock
        percentage_spy: Mock

        @pytest.fixture(autouse=True, scope='function')
        def patch(self, mocker: MockerFixture):
            self.setter_spy = mocker.spy(sbc, 'set_brightness')
            self.percentage_spy = mocker.spy(sbc, 'percentage')

        def test_relative_values_are_calculated(self, mocker: MockerFixture):
            mocker.patch.object(sbc.Display, 'get_brightness', new=lambda *a, **k: 10)
            display = sbc.Display.from_dict(sbc.list_monitors_info()[0])
            spy = mocker.spy(display.method, 'set_brightness')

            sbc.set_brightness('+5', display=0)
            # check `percentage` is called and returns the correct value
            assert self.percentage_spy.spy_return == 15
            # check the result is passed to `set_brightness()` of `BrightnessMethod`
            spy.assert_called_once_with(15, display=display.index)

        def test_current_value_if_get_brightness_fails(self, mocker: MockerFixture):
            '''
            For relative brightnesses, we need to fetch the current brightness and add the relative
            value to it. If `get_brightness` returns None (ie: fails) then we need a fallback behaviour
            '''
            mocker.patch.object(sbc, 'get_brightness', new=lambda *a, **k: [None])
            sbc.set_brightness('+10', display=0)
            assert self.percentage_spy.mock_calls[0].kwargs.get('current') is not None

        def test_relative_values_are_per_display(self, mocker: MockerFixture):
            count = -1

            def counter(*a, **kw):
                nonlocal count
                count += 1
                return [count]

            mocker.patch.object(sbc, 'get_brightness', new=counter)
            sbc.set_brightness('+10')
            expected = [i + 10 for i in range(count + 1)]
            actual = [call.args[0] for call in self.setter_spy.mock_calls[1:]]
            assert expected == actual


class TestFadeBrightness(BrightnessFunctionTest):
    @pytest.fixture
    def operation_type(self):
        return 'fade'

    def test_returns_new_brightness_by_default(self, displays):
        result = sbc.fade_brightness(100, interval=0)
        assert isinstance(result, list) and all(isinstance(i, int) for i in result)
        # `type: ignore` because fade brightness could return `list[Thread]`
        assert sorted(result) == sorted(d['index'] for d in displays)  # type: ignore

    def test_blocking_kwarg(self, subtests):
        threads = sbc.fade_brightness(100, blocking=False, interval=0)
        assert isinstance(threads, list) and all(isinstance(t, threading.Thread) for t in threads)
        for index, thread in enumerate(threads):
            with subtests.test(index=index, thread=thread):
                # assert again for type checker
                assert isinstance(thread, threading.Thread)
                thread.join()

    def test_passes_kwargs_to_display_class(self, mocker: MockerFixture, subtests):
        '''
        Most of the fade logic has been moved to `Display.fade_brightness`. The top level
        `fade_brightness` function is just responsible for coordinating all the different displays.

        This test just checks that we pass all the correct config to the display class, and then the
        `Display` unit tests will check that all the right things happen
        '''

        def stub(*a, **k):
            pass

        spy = mocker.patch.object(sbc.Display, 'fade_brightness', Mock(side_effect=stub))
        args = (100,)
        # all the kwargs that get passed to `Display`
        kwargs: Dict[str, Any] = dict(start=0, interval=0, increment=10, force=False, logarithmic=False)
        sbc.fade_brightness(*args, **kwargs)
        for index, mock_call in enumerate(spy.mock_calls):
            with subtests.test(index=index, mock_call=mock_call):
                assert mock_call == call(*args, **kwargs)


def test_list_monitors_info(mock_os_module, mocker: MockerFixture):
    '''
    `list_monitors_info` is just a shell for the OS specific variant
    '''
    mock = mocker.patch.object(sbc._OS_MODULE, 'list_monitors_info', Mock(return_value=12345, spec=True))
    supported_kw = {'method': 123, 'allow_duplicates': 456, 'unsupported': 789}
    result = sbc.list_monitors_info(**supported_kw)  # type: ignore
    # check that kwargs passed along and result passed back
    mock.assert_called_once_with(**supported_kw)
    assert result == 12345


def test_list_monitors(mock_os_module, mocker: MockerFixture):
    '''
    `list_monitors` is just a shell for `list_monitors_info`
    '''
    mock_return = [{'name': '123'}, {'name': '456'}]
    mock = mocker.patch.object(sbc, 'list_monitors_info', Mock(return_value=mock_return, spec=True))
    supported_kw = {'method': 123, 'allow_duplicates': 456}
    result = sbc.list_monitors(**supported_kw)  # type:ignore
    # check that kwargs passed along and result passed back
    mock.assert_called_once_with(**supported_kw)
    assert result == [i['name'] for i in mock_return]


class TestGetMethods:
    def test_returns_dict_of_brightness_methods(self, subtests):
        methods = sbc.get_methods()
        assert isinstance(methods, dict)
        # check all methods included
        assert tuple(methods.values()) == sbc._OS_MODULE.METHODS
        # check names match up
        for name, method_class in methods.items():
            with subtests.test(method=name):
                assert name == method_class.__name__.lower()
                assert issubclass(method_class, sbc.BrightnessMethod)

    class TestNameKwarg:
        def test_non_str_raises_type_error(self):
            with pytest.raises(TypeError, match=r'name must be of type str.*'):
                sbc.get_methods(sbc._OS_MODULE.METHODS[0])  # type: ignore

        def test_raises_value_error_on_invalid_lookup(self):
            with pytest.raises(ValueError, match=r'invalid method.*'):
                sbc.get_methods('does not exist')

        @pytest.mark.parametrize('name,method_class', [(i.__name__.lower(), i) for i in os_module_mock.METHODS])
        def test_returns_dict_on_valid_lookup(self, mock_os_module, name: str, method_class):
            assert sbc.get_methods(name) == {name: method_class}

        @pytest.mark.parametrize('name,method_class', [(i.__name__.upper(), i) for i in os_module_mock.METHODS])
        def test_converts_lookups_to_lowercase(self, mock_os_module, name: str, method_class):
            assert sbc.get_methods(name) == {name.lower(): method_class}


class TestDisplay:
    @pytest.fixture(autouse=True, scope='function')
    def display(self) -> sbc.Display:
        '''Returns a `Display` instance with the brightness set to 50'''
        display = sbc.Display.from_dict(sbc.list_monitors_info()[0])
        display.set_brightness(50)
        return display

    class TestFadeBrightness:
        @pytest.mark.parametrize('value', [100, 0, 75, 50, 150, -10])
        def test_returns_none(self, display: sbc.Display, value: int):
            assert display.fade_brightness(value, interval=0) is None

        @pytest.mark.parametrize('value', ['60', '70.0', '+10', '-10', '500'])
        def test_relative_values(self, display: sbc.Display, value):
            display.fade_brightness(value, interval=0)
            assert display.get_brightness() == sbc.percentage(value, current=50)

        @pytest.mark.parametrize('value', [100, 75, 50, 25])
        def test_start_kwarg(self, display: sbc.Display, mocker: MockerFixture, value):
            spy = mocker.spy(display, 'set_brightness')
            display.fade_brightness(100, start=value, interval=0)
            assert spy.mock_calls[0].args[0] == value

        def test_interval_kwarg(self, display: sbc.Display):
            assert timeit(lambda: display.fade_brightness(100, start=95, interval=0), number=1) < timeit(
                lambda: display.fade_brightness(100, start=95, interval=0.05), number=1
            ), 'longer interval should take more time'

        @pytest.mark.parametrize('increment', [1, 5, 10, 15])
        @pytest.mark.parametrize('start', [0, 100])
        def test_increment_kwarg(self, display: sbc.Display, mocker: MockerFixture, increment: int, start: int):
            target = 50
            spy = mocker.spy(display, 'set_brightness')
            display.fade_brightness(target, interval=0, increment=increment, logarithmic=False, start=start)
            values = [call.args[0] for call in spy.mock_calls]
            # go until len - 2 because the last call to `set_brightness` is usually to make up the
            # difference between the last incremented step and the target value
            diffs = [values[i + 1] - values[i] for i in range(len(values) - 2)]

            # check that it works the same when fading to a dimmer value
            if start > target:
                increment = -increment
            assert set(diffs) == {increment}

        @pytest.mark.parametrize('os_name', ['Windows', 'Linux'])
        def test_force_kwarg(self, display: sbc.Display, mocker: MockerFixture, os_name: str):
            mocker.patch.object(sbc.platform, 'system', new=lambda: os_name)
            lower_bound = 1 if os_name == 'Linux' else 0
            spy = mocker.spy(display, 'set_brightness')

            display.fade_brightness(10, start=0, interval=0)
            assert spy.mock_calls[0].args[0] == lower_bound
            spy.reset_mock()

            display.fade_brightness(10, start=0, interval=0, force=True)
            assert spy.mock_calls[0].args[0] == 0

        def test_logarithmic_kwarg(self, display: sbc.Display, mocker: MockerFixture):
            # range_spy = mocker.spy(sbc, 'range')  # cant spy on range?
            logarithmic_range_spy = mocker.spy(sbc, 'logarithmic_range')

            display.fade_brightness(100, interval=0)
            # range_spy.assert_not_called()
            logarithmic_range_spy.assert_called()

            # range_spy.reset_mock()
            logarithmic_range_spy.reset_mock()

            display.fade_brightness(100, interval=0, logarithmic=False)
            # range_spy.assert_called()
            logarithmic_range_spy.assert_not_called()

        def test_end_of_fade_correction(self, display: sbc.Display, mocker: MockerFixture):
            '''
            If the brightness does not match the target at the end of the fade then
            this should be corrected
            '''
            target = 100
            mocker.patch.object(display, 'get_brightness', Mock(return_value=50))
            setter = mocker.patch.object(display, 'set_brightness', autospec=True)
            # patch the range function so that it never returns the target brightness
            range_values = list(sbc.logarithmic_range(0, 100))[:-10]
            assert target not in range_values, 'setup has gone wrong!'
            mocker.patch.object(sbc, 'logarithmic_range', Mock(return_value=range_values, spec=True))

            display.fade_brightness(target, start=0, interval=0)
            # at this point, fade_brightness should have manually set the final brightness
            assert setter.mock_calls[-1].args[0] == target
            # it should have also passed the `force` kwarg along to the final call
            assert 'force' in setter.mock_calls[-1].kwargs, 'force kwarg should be propagated'

        def test_stoppable_kwarg(self, display: sbc.Display, mocker: MockerFixture):
            start = 1
            finish = 20  # smaller value could introduce errors; greater value will extend the test.
            interval = 0.05  # same as above
            steps = int((finish - start) / 2)  # half the steps to ensure the thread is still active when checking.
            duration = interval * (steps - 1)  # -1 because the first step is immediate.

            mocker.patch.object(display, 'get_brightness', Mock(return_value=start))
            setter = mocker.patch.object(display, 'set_brightness', autospec=True)

            def fade_brightness_thread(stoppable: bool):
                '''mainly for Mypy to stop complaining about the return type of `display.fade_brightness`'''
                return cast(
                    threading.Thread,
                    display.fade_brightness(
                        finish=finish,
                        start=start,
                        interval=interval,
                        logarithmic=False,
                        blocking=False,
                        stoppable=stoppable,
                    ),
                )

            thread_0 = fade_brightness_thread(stoppable=True)
            thread_1 = fade_brightness_thread(stoppable=True)
            time.sleep(duration)  # block the main thread to allow non-blocking fades to occur.
            # The second fade should have stopped the first one.
            assert not thread_0.is_alive() and thread_1.is_alive()
            call_count = len(setter.mock_calls)
            assert round(call_count / steps) == 1  # 1 stands for the expected number of running threads

            # The fades below can't be stopped but they will halt the two above, which is essential for call count.
            thread_2 = fade_brightness_thread(stoppable=False)
            thread_3 = fade_brightness_thread(stoppable=False)
            time.sleep(duration)
            # Both two new threads should run without stopping.
            assert thread_2.is_alive() and thread_3.is_alive()
            call_count = len(setter.mock_calls) - call_count
            assert round(call_count / steps) == 2  # 2 stands for the expected number of running threads

            # Ensure all threads complete to prevent interference with subsequent tests.
            while threading.active_count() > 1:
                time.sleep(interval)

    class TestFromDict:
        def test_returns_valid_instance(self, subtests):
            info = sbc.list_monitors_info()[0]
            display = sbc.Display.from_dict(info)
            assert isinstance(display, sbc.Display)
            for field in dataclasses.fields(sbc.Display):
                if field.name.startswith('_'):
                    continue
                with subtests.test(field=field):
                    assert getattr(display, field.name) == info[field.name]

        def test_excludes_extra_fields(self):
            info = {**sbc.list_monitors_info()[0], 'extra': '12345'}
            display = sbc.Display.from_dict(info)
            with pytest.raises(AttributeError):
                getattr(display, 'extra')

    def test_get_brightness(self, display: sbc.Display, mocker: MockerFixture):
        spy = mocker.spy(display.method, 'get_brightness')
        result = display.get_brightness()
        spy.assert_called_once_with(display=display.index)
        # method returns list[int]. display should return int
        assert isinstance(result, int) and result == spy.spy_return[0]

    class TestGetIdentifier:
        def test_returns_tuple(self, display: sbc.Display):
            result = display.get_identifier()
            assert isinstance(result, tuple)
            prop, value = result
            assert isinstance(prop, str) and hasattr(display, prop)
            assert getattr(display, prop) == value
            assert value is not None

        @pytest.mark.parametrize('prop', ['uid', 'edid', 'serial', 'name', 'index'])
        def test_returns_first_not_none_value(self, display: sbc.Display, prop: str):
            all_props = ['uid', 'edid', 'serial', 'name', 'index']
            for p in all_props:
                if p == prop:
                    continue
                setattr(display, p, None)

            key, value = display.get_identifier()
            assert key == prop and value == getattr(display, prop)

        def test_allows_falsey_values(self, display: sbc.Display):
            '''
            `get_identifier` should skip properties only if they are `None`.
            It's very easy to do an `if truthy` check but that's not what we want here.
            '''
            display.uid = ''
            prop, value = display.get_identifier()
            assert prop == 'uid' and value == ''

    def test_is_active(self, display: sbc.Display, mocker: MockerFixture):
        # normal operation, should return true
        assert display.is_active() is True

        def stub(*_, **__):
            raise Exception

        mocker.patch.object(display, 'get_brightness', Mock(side_effect=stub))
        # if get_brightness fails, should return false
        assert display.is_active() is False

    class TestSetBrightness:
        @pytest.mark.parametrize('value', [1, 10, 21, 37, 43, 50, 90, 100])
        def test_normal(self, display: sbc.Display, mocker: MockerFixture, value: int):
            spy = mocker.spy(display.method, 'set_brightness')
            display.set_brightness(value)
            spy.assert_called_once_with(value, display=display.index)

        def test_returns_none(self):
            assert sbc.set_brightness(100) is None

        def test_relative_values(self, display: sbc.Display, mocker: MockerFixture):
            mocker.patch.object(display, 'get_brightness', Mock(return_value=50))
            spy = mocker.spy(display.method, 'set_brightness')
            display.set_brightness('+30')
            spy.assert_called_once_with(80, display=display.index)

        @pytest.mark.parametrize('os_name', ['Windows', 'Linux'])
        def test_force_kwarg(self, display: sbc.Display, mocker: MockerFixture, os_name: str):
            mocker.patch.object(sbc.platform, 'system', new=lambda: os_name)
            lower_bound = 1 if os_name == 'Linux' else 0
            spy = mocker.spy(display.method, 'set_brightness')

            display.set_brightness(0)
            assert spy.mock_calls[0].args[0] == lower_bound
            spy.reset_mock()

            display.set_brightness(0, force=True)
            assert spy.mock_calls[0].args[0] == 0


class TestFilterMonitors:
    def test_returns_list_of_dict(self):
        filtered = sbc.filter_monitors()
        assert isinstance(filtered, list)
        assert all(isinstance(i, dict) for i in filtered)

    def test_raises_exception_when_no_displays_detected(self, mocker: MockerFixture):
        mocker.patch.object(sbc, 'list_monitors_info', Mock(spec=True, return_value=[]))
        # filter_monitors sleeps 0.4s between retries. patch that to speed up tests
        mocker.patch.object(sbc.time, 'sleep', Mock())
        with pytest.raises(sbc.NoValidDisplayError):
            sbc.filter_monitors()

    class TestDisplayKwarg:
        sample_monitors: List[dict]

        @pytest.fixture(autouse=True, scope='function')
        def setup(self, mocker: MockerFixture):
            methods = tuple(sbc.get_methods().values())
            self.sample_monitors = [
                {'index': 0, 'method': methods[0], 'name': 'Dell Sample 1', 'serial': '1234', 'edid': '00ffwhatever'},
                {
                    'index': 1,
                    'method': methods[0],
                    # duplicate of sample 1
                    'name': 'Dell Sample 1',
                    'serial': '1234',
                    'edid': '00ffwhatever',
                },
                {'index': 0, 'method': methods[1], 'name': 'Dell Sample 2'},
            ]
            mocker.patch.object(sbc, 'list_monitors_info', Mock(spec=True, return_value=self.sample_monitors))
            return self.sample_monitors

        @pytest.mark.parametrize('invalid_input', [[], 0.0])
        def test_raises_type_error_on_invalid_display_kwarg(self, invalid_input):
            with pytest.raises(TypeError):
                sbc.filter_monitors(display=invalid_input)

        def test_filters_displays_by_global_index(self):
            result = sbc.filter_monitors(display=0)
            assert len(result) == 1 and result[0] == self.sample_monitors[0]

            # Test filtering duplicates with different parameters
            # - When 'allow_duplicates' parameter is not set (using default value False), the duplicate is filtered out
            assert len(sbc.filter_monitors()) == 2
            assert sbc.filter_monitors(display=1) == [self.sample_monitors[2]]

            # - When 'allow_duplicates' is set to True, the duplicate is preserved
            assert len(sbc.filter_monitors(allow_duplicates=True)) == 3
            assert sbc.filter_monitors(display=1, allow_duplicates=True) == [self.sample_monitors[1]]

            # - When 'allow_duplicates' is set to False, the duplicate is filtered out
            assert len(sbc.filter_monitors(allow_duplicates=False)) == 2
            assert sbc.filter_monitors(display=1, allow_duplicates=False) == [self.sample_monitors[2]]

        class TestDuplicateFilteringAndIncludeKwarg:
            default_identifiers = ['edid', 'serial', 'name']

            @pytest.fixture(scope='function')
            def sample_monitors(self, setup):
                return deepcopy(setup[:2])

            @pytest.mark.parametrize('field', default_identifiers)
            def test_filters_duplicates_by_first_not_none_identifier(
                self, sample_monitors: List[dict], field: str, include=None
            ):
                '''
                There are 3 properties of a display we can use to identify it: edid, serial and name.
                EDID contains the serial and is the most unique. Two monitors with the same edid are
                the same display, but we can't always get the EDID (laptop displays). Serial is also
                pretty unique, but again we can't always get it.
                The name is only somewhat unique due to the fact that I rarely see someone with two
                of the same display. It's always a primary and some random.

                We should prioritize these identifiers in terms of uniqueness, with edid first and name
                last. If one is not available, fall back to the next one.
                '''
                include = include or []
                identifier_fields = deepcopy(self.default_identifiers) + include
                for item in sample_monitors:
                    # delete all identifier fields that take priority over this one
                    for f in identifier_fields:
                        if f == field:
                            break
                        del item[f]

                assert sbc.filter_monitors(haystack=sample_monitors, include=include) == [sample_monitors[0]], (
                    f'both displays have same {field!r}, second should be filtered out'
                )

                sample_monitors[0][field] = str(reversed(sample_monitors[1][field]))
                assert sbc.filter_monitors(haystack=sample_monitors, include=include) == sample_monitors, (
                    f'both displays have different {field!r}s, neither should be filtered out'
                )

                del sample_monitors[0][field]
                if field == identifier_fields[-1]:
                    # special case for 'name' because this is the last valid identifier
                    assert sbc.filter_monitors(haystack=sample_monitors, include=include) == [sample_monitors[1]], (
                        'first display has no valid identifiers and should be removed'
                    )
                else:
                    assert sbc.filter_monitors(haystack=sample_monitors, include=include) == sample_monitors, (
                        f'one display is missing {field!r}, neither should be filtered'
                    )

                del sample_monitors[1][field]
                if field == identifier_fields[-1]:
                    with pytest.raises(sbc.NoValidDisplayError):
                        # neither display has any valid identifiers. Both should be removed
                        sbc.filter_monitors(haystack=sample_monitors, include=include)
                else:
                    assert sbc.filter_monitors(haystack=sample_monitors, include=include) == [sample_monitors[0]], (
                        f'neither display has {field!r}, second should be filtered out'
                    )

            @pytest.mark.parametrize('field', [default_identifiers[-1], 'extra'])
            def test_include_kwarg_acts_as_identifier_when_filtering_duplicates(
                self, sample_monitors: List[dict], field: str
            ):
                for item in sample_monitors:
                    item['extra'] = '12345'

                self.test_filters_duplicates_by_first_not_none_identifier(sample_monitors, field, [field])

            def test_include_kwarg_can_identify_displays(self, sample_monitors: List[dict]):
                for item in sample_monitors:
                    for field in self.default_identifiers:
                        del item[field]

                sample_monitors[0]['extra'] = 'extra_info'
                with pytest.raises(sbc.NoValidDisplayError):
                    # it shouldn't work without the include
                    sbc.filter_monitors(display='extra_info', haystack=sample_monitors)

                assert sbc.filter_monitors(display='extra_info', include=['extra'], haystack=sample_monitors) == [
                    sample_monitors[0]
                ]

            def test_identifiers_that_do_not_match_display_kwarg_are_not_used(self):
                '''
                If two displays have the same edid but different names and we search for
                the second name, it should return the second display. It should not filter it out
                as a duplicate of the first.
                '''
                primary = sbc.list_monitors_info()[0]
                displays = [deepcopy(primary) for i in range(3)]
                displays[-1]['name'] = 'Display with weird name'
                assert sbc.filter_monitors(haystack=displays) == [displays[0]], (
                    'no display kwarg, duplciates are filtered on edid'
                )
                assert sbc.filter_monitors(haystack=displays, display=displays[-1]['name']) == [displays[-1]], (
                    'display kwarg present, displays are filtered by whaever identifier matches'
                )

    class TestHaystackAndMethodKwargs:
        class TestWithHaystack:
            def test_skips_calling_list_monitors_info(self, mocker: MockerFixture):
                displays = sbc.list_monitors_info()
                spy = mocker.spy(sbc, 'list_monitors_info')
                sbc.filter_monitors(haystack=displays)
                spy.assert_not_called()

            def test_filters_by_method(self):
                method_name, method_class = next(iter(sbc.get_methods().items()))
                result = sbc.filter_monitors(method=method_name, haystack=sbc.list_monitors_info())
                assert all(display['method'] == method_class for display in result)

            def test_does_not_mutate_the_haystack(self):
                haystack = sbc.list_monitors_info()
                haystack_orig = deepcopy(haystack)
                sbc.filter_monitors(method=haystack[0]['method'].__name__, haystack=haystack)
                assert haystack == haystack_orig

        class TestWithoutHaystack:
            def test_filters_from_list_with_duplicates(self, mocker: MockerFixture):
                '''It should allow duplictaes to be returned from `list_monitors_info` and filter the whole list'''
                spy = mocker.spy(sbc, 'list_monitors_info')
                sbc.filter_monitors()
                spy.assert_called()
                assert spy.mock_calls[0].kwargs.get('allow_duplicates', False) is True

            def test_passes_method_kwarg_along(self, mocker: MockerFixture):
                spy = mocker.spy(sbc, 'list_monitors_info')
                method = next(iter(sbc.get_methods().keys()))
                sbc.filter_monitors(method=method)
                spy.assert_called()
                assert spy.mock_calls[0].kwargs.get('method', None) == method

        @pytest.mark.parametrize('haystack', [None, []])
        def test_error_raised_on_invalid_method_kwarg(self, haystack):
            with pytest.raises(ValueError):
                sbc.filter_monitors(method='not real method', haystack=haystack)



from pytest import MonkeyPatch

from screen_brightness_control import config


def test_default_params(monkeypatch: MonkeyPatch):
    func = config.default_params(lambda **kw: kw)

    assert func() == {'allow_duplicates': config.ALLOW_DUPLICATES, 'method': config.METHOD}, (
        'sets default kwarg values'
    )

    monkeypatch.setattr(config, 'METHOD', 'my_method')
    monkeypatch.setattr(config, 'ALLOW_DUPLICATES', 123)
    assert func() == {'allow_duplicates': 123, 'method': 'my_method'}

    assert func(method=None, allow_duplicates=None) == {'method': None, 'allow_duplicates': None}, (
        'should not override kwargs'
    )



import ctypes
from contextlib import contextmanager
from typing import Type
from unittest.mock import Mock, call

import pytest
from pytest_mock import MockerFixture

import screen_brightness_control as sbc
from screen_brightness_control.helpers import BrightnessMethod

from .helpers import BrightnessMethodTest
from .mocks.windows_mock import FakeWinDLL, mock_enum_display_devices, mock_enum_display_monitors, mock_wmi_init


@pytest.fixture
def patch_global_get_display_info(mocker: MockerFixture):
    '''Mock everything needed to get `sbc.windows.get_display_info` to run'''
    mocker.patch.object(sbc.windows, 'enum_display_devices', mock_enum_display_devices)
    mocker.patch.object(sbc.windows, '_wmi_init', mock_wmi_init)
    mocker.patch.object(sbc.windows.win32api, 'EnumDisplayMonitors', mock_enum_display_monitors)


class TestWMI(BrightnessMethodTest):
    @pytest.fixture
    def patch_get_display_info(self, patch_global_get_display_info):
        '''Mock everything needed to get `WMI.get_display_info` to run'''
        pass

    @pytest.fixture
    def patch_get_brightness(self, mocker: MockerFixture, patch_get_display_info):
        pass

    @pytest.fixture
    def patch_set_brightness(self, mocker: MockerFixture, patch_get_display_info):
        pass

    @pytest.fixture
    def method(self) -> Type[BrightnessMethod]:
        return sbc.windows.WMI

    class TestGetBrightness(BrightnessMethodTest.TestGetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestGetBrightness.TestDisplayKwarg):
            # skip these because WMI doesn't really make display specific calls when getting brightness
            # and perf is negligible anyway
            @pytest.mark.skip('skip TestGetBrightness.TestDisplayKwarg perf tests for WMI')
            def test_with(self):
                pass

            @pytest.mark.skip('skip TestGetBrightness.TestDisplayKwarg perf tests for WMI')
            def test_without(self):
                pass

    class TestSetBrightness(BrightnessMethodTest.TestSetBrightness):
        class TestDisplayKwarg(BrightnessMethodTest.TestSetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, freeze_display_info, method, subtests):
                with sbc.windows._wmi_init() as wmi:
                    mocker.patch.object(
                        sbc.windows, '_wmi_init', Mock(side_effect=contextmanager(lambda *_: (yield wmi)))
                    )
                    brightness_method = wmi.WmiMonitorBrightnessMethods()[0]
                    mocker.patch.object(wmi, 'WmiMonitorBrightnessMethods', lambda: [brightness_method] * 3)
                    spy = mocker.spy(brightness_method, 'WmiSetBrightness')
                    for index, display in enumerate(freeze_display_info):
                        with subtests.test(index=index):
                            method.set_brightness(100, display=index)
                            spy.assert_called_once_with(100, 0)
                            spy.reset_mock()

            def test_without(self, mocker: MockerFixture, freeze_display_info, method):
                with sbc.windows._wmi_init() as wmi:
                    mocker.patch.object(
                        sbc.windows, '_wmi_init', Mock(side_effect=contextmanager(lambda *_: (yield wmi)))
                    )
                    brightness_method = wmi.WmiMonitorBrightnessMethods()[0]
                    mocker.patch.object(wmi, 'WmiMonitorBrightnessMethods', lambda: [brightness_method] * 3)
                    spy = mocker.spy(brightness_method, 'WmiSetBrightness')

                    method.set_brightness(100)
                    spy.assert_has_calls([call(100, 0)] * 3)
                    spy.reset_mock()


class TestVCP(BrightnessMethodTest):
    @pytest.fixture
    def patch_get_display_info(self, patch_global_get_display_info, mocker: MockerFixture, method):
        '''Mock everything needed to get `VCP.get_display_info` to run'''
        mocker.patch.object(ctypes, 'windll', FakeWinDLL, create=True)
        # also patch locally imported version
        mocker.patch.object(sbc.windows, 'windll', FakeWinDLL)

        sbc.windows.__cache__.enabled = False

    @pytest.fixture
    def patch_get_brightness(self, mocker: MockerFixture, patch_get_display_info):
        pass

    @pytest.fixture
    def patch_set_brightness(self, mocker: MockerFixture, patch_get_display_info):
        pass

    @pytest.fixture
    def method(self) -> Type[BrightnessMethod]:
        return sbc.windows.VCP

    class TestGetBrightness(BrightnessMethodTest.TestGetBrightness):
        @pytest.mark.parametrize('display_kw', [None, 0, 1, 2])
        def test_handles_are_cleaned_up(self, mocker: MockerFixture, method, display_kw):
            '''
            DestroyPhysicalMonitor should be called ONCE for each monitor iterated over.
            When using the `display` kwarg, we still have to iterate over those displays in
            `iter_physical_monitors`, but we only yield from the `start` kw, meaning that
            `n + 1` handles must be destroyed, where n is the target display index.
            '''
            num_displays = display_kw + 1 if display_kw is not None else len(method.get_display_info())
            spy = mocker.spy(ctypes.windll.dxva2, 'DestroyPhysicalMonitor')
            method.get_brightness(display=display_kw)
            assert spy.call_count == num_displays

        # 0 is included here to make sure loop variables aren't referenced before assignment
        @pytest.mark.parametrize('tries', [0, 1, 2, 3, 50, 100])
        def test_max_tries(self, mocker: MockerFixture, method, tries):
            mock = mocker.patch.object(
                sbc.windows.windll.dxva2, 'GetVCPFeatureAndVCPFeatureReply', Mock(return_value=0, spec=True)
            )
            # patch sleep func so tests aren't slowed down
            mocker.patch.object(sbc.windows.time, 'sleep', Mock(auto=True))
            method.get_brightness(display=0, max_tries=tries)
            assert mock.call_count == tries

        class TestDisplayKwarg(BrightnessMethodTest.TestGetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, freeze_display_info, method, subtests):
                spy = mocker.spy(ctypes.windll.dxva2, 'GetVCPFeatureAndVCPFeatureReply')
                handles = tuple(sbc.windows.VCP.iter_physical_monitors())
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.get_brightness(display=index)
                        spy.assert_called_once()
                        assert spy.mock_calls[0].args[0] == handles[index]
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method, subtests):
                spy = mocker.spy(ctypes.windll.dxva2, 'GetVCPFeatureAndVCPFeatureReply')
                handles = tuple(sbc.windows.VCP.iter_physical_monitors())
                method.get_brightness()
                spy.assert_called()
                for index, handle in enumerate(handles):
                    with subtests.test(index=index):
                        assert spy.mock_calls[index].args[0] == handle

    class TestSetBrightness(BrightnessMethodTest.TestSetBrightness):
        @pytest.mark.parametrize('display_kw', [None, 0, 1, 2])
        def test_handles_are_cleaned_up(self, mocker: MockerFixture, method, display_kw):
            '''
            DestroyPhysicalMonitor should be called ONCE for each monitor iterated over.
            When using the `display` kwarg, we still have to iterate over those displays in
            `iter_physical_monitors`, but we only yield from the `start` kw, meaning that
            `n + 1` handles must be destroyed, where n is the target display index.
            '''
            num_displays = display_kw + 1 if display_kw is not None else len(method.get_display_info())
            spy = mocker.spy(ctypes.windll.dxva2, 'DestroyPhysicalMonitor')
            method.set_brightness(100, display=display_kw)
            assert spy.call_count == num_displays

        # 0 is included here to make sure loop variables aren't referenced before assignment
        @pytest.mark.parametrize('tries', [0, 1, 2, 3, 50, 100])
        def test_max_tries(self, mocker: MockerFixture, method, tries):
            mock = mocker.patch.object(sbc.windows.windll.dxva2, 'SetVCPFeature', Mock(return_value=0, spec=True))
            # patch sleep func so tests aren't slowed down
            mocker.patch.object(sbc.windows.time, 'sleep', Mock(auto=True))
            method.set_brightness(100, display=0, max_tries=tries)
            assert mock.call_count == tries

        class TestDisplayKwarg(BrightnessMethodTest.TestSetBrightness.TestDisplayKwarg):
            def test_with(self, mocker: MockerFixture, freeze_display_info, method, subtests):
                spy = mocker.spy(ctypes.windll.dxva2, 'SetVCPFeature')
                handles = tuple(sbc.windows.VCP.iter_physical_monitors())
                for index, display in enumerate(freeze_display_info):
                    with subtests.test(index=index):
                        method.set_brightness(100, display=index)
                        spy.assert_called_once()
                        assert spy.mock_calls[0].args[0] == handles[index]
                        spy.reset_mock()

            def test_without(self, mocker: MockerFixture, method, subtests):
                spy = mocker.spy(ctypes.windll.dxva2, 'SetVCPFeature')
                handles = tuple(sbc.windows.VCP.iter_physical_monitors())
                method.set_brightness(100)
                spy.assert_called()
                for index, handle in enumerate(handles):
                    with subtests.test(index=index):
                        assert spy.mock_calls[index].args[0] == handle



import fcntl
import functools
import glob
import logging
import operator
import os
import re
import time
from typing import List, Optional, Tuple
import warnings

from . import filter_monitors, get_methods
from .exceptions import I2CValidationError, NoValidDisplayError, format_exc
from .helpers import EDID, BrightnessMethod, BrightnessMethodAdv, __Cache, _monitor_brand_lookup, check_output
from .types import DisplayIdentifier, IntPercentage

__cache__ = __Cache()
_logger = logging.getLogger(__name__)


class SysFiles(BrightnessMethod):
    '''
    A way of getting display information and adjusting the brightness
    that does not rely on any 3rd party software.

    This class works with displays that show up in the `/sys/class/backlight`
    directory (so usually laptop displays).

    To set the brightness, your user will need write permissions for
    `/sys/class/backlight/*/brightness` or you will need to run the program
    as root.
    '''

    _logger = _logger.getChild('SysFiles')

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        subsystems = set()
        for folder in os.listdir('/sys/class/backlight'):
            if os.path.isdir(f'/sys/class/backlight/{folder}/subsystem'):
                subsystems.add(tuple(os.listdir(f'/sys/class/backlight/{folder}/subsystem')))

        displays_by_edid = {}
        index = 0

        # map drm devices to their pci device paths
        drm_paths = {}
        for folder in glob.glob('/sys/class/drm/card*-*'):
            drm_paths[os.path.realpath(folder)] = folder

        for subsystem in subsystems:
            device: dict = {
                'name': subsystem[0],
                'path': f'/sys/class/backlight/{subsystem[0]}',
                'method': cls,
                'index': index,
                'model': None,
                'serial': None,
                'manufacturer': None,
                'manufacturer_id': None,
                'edid': None,
                'scale': None,
                'uid': None,
            }

            for folder in subsystem:
                # subsystems like intel_backlight usually have an acpi_video0
                # counterpart, which we don't want so lets find the 'best' candidate
                try:
                    with open(f'/sys/class/backlight/{folder}/max_brightness') as f:
                        # scale for SysFiles is just a multiplier for the set/get brightness values
                        scale = int(f.read().rstrip(' \n')) / 100

                    # use the display with the highest resolution scale
                    if device['scale'] is None or scale > device['scale']:
                        device['name'] = folder
                        device['path'] = f'/sys/class/backlight/{folder}'
                        device['scale'] = scale
                except (FileNotFoundError, TypeError) as e:
                    cls._logger.error(f'error getting highest resolution scale for {folder} - {format_exc(e)}')
                    continue

                # check if backlight subsystem device matches any of the PCI devices discovered earlier
                # if so, extract the i2c bus from the drm device folder
                pci_path = os.path.realpath(f'/sys/class/backlight/{folder}/device')
                if pci_path in drm_paths:
                    device['uid'] = device['uid'] or i2c_bus_from_drm_device(drm_paths[pci_path])

            if os.path.isfile('%s/device/edid' % device['path']):
                device['edid'] = EDID.hexdump('%s/device/edid' % device['path']) or None

                if device['edid'] is None:
                    cls._logger.warning(f'EDID file exists but is empty for display {device["path"]}')
                else:
                    for key, value in zip(
                        ('manufacturer_id', 'manufacturer', 'model', 'name', 'serial'), EDID.parse(device['edid'])
                    ):
                        if value is None:
                            continue
                        device[key] = value

            displays_by_edid[device['edid']] = device
            index += 1

        all_displays = list(displays_by_edid.values())
        if display is not None:
            all_displays = filter_monitors(display=display, haystack=all_displays, include=['path'])
        return all_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        results = []
        for device in info:
            with open(os.path.join(device['path'], 'brightness'), 'r') as f:
                brightness = int(f.read().rstrip('\n'))
            results.append(int(brightness / device['scale']))

        return results

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        for device in info:
            with open(os.path.join(device['path'], 'brightness'), 'w') as f:
                f.write(str(int(value * device['scale'])))


class I2C(BrightnessMethod):
    '''
    In the same spirit as `SysFiles`, this class serves as a way of getting
    display information and adjusting the brightness without relying on any
    3rd party software.

    Usage of this class requires read and write permission for `/dev/i2c-*`.

    This class works over the I2C bus, primarily with desktop monitors as I
    haven't tested any e-DP displays yet.

    Massive thanks to [siemer](https://github.com/siemer) for
    his work on the [ddcci.py](https://github.com/siemer/ddcci) project,
    which served as a my main reference for this.

    References:
        * [ddcci.py](https://github.com/siemer/ddcci)
        * [DDCCI Spec](https://milek7.pl/ddcbacklight/ddcci.pdf)
    '''

    _logger = _logger.getChild('I2C')

    # vcp commands
    GET_VCP_CMD = 0x01
    '''VCP command to get the value of a feature (eg: brightness)'''
    GET_VCP_REPLY = 0x02
    '''VCP feature reply op code'''
    SET_VCP_CMD = 0x03
    '''VCP command to set the value of a feature (eg: brightness)'''

    # addresses
    DDCCI_ADDR = 0x37
    '''DDC packets are transmittred using this I2C address'''
    HOST_ADDR_R = 0x50
    '''Packet source address (the computer) when reading data'''
    HOST_ADDR_W = 0x51
    '''Packet source address (the computer) when writing data'''
    DESTINATION_ADDR_W = 0x6E
    '''Packet destination address (the monitor) when writing data'''
    I2C_SLAVE = 0x0703
    '''The I2C slave address'''

    # timings
    WAIT_TIME = 0.05
    '''How long to wait between I2C commands'''

    _max_brightness_cache: dict = {}

    class I2CDevice:
        '''
        Class to read and write data to an I2C bus,
        based on the `I2CDev` class from [ddcci.py](https://github.com/siemer/ddcci)
        '''

        def __init__(self, fname: str, slave_addr: int):
            '''
            Args:
                fname: the I2C path, eg: `/dev/i2c-2`
                slave_addr: not entirely sure what this is meant to be
            '''
            self.device = os.open(fname, os.O_RDWR)
            # I2C_SLAVE address setup
            fcntl.ioctl(self.device, I2C.I2C_SLAVE, slave_addr)

        def read(self, length: int) -> bytes:
            '''
            Read a certain number of bytes from the I2C bus

            Args:
                length: the number of bytes to read

            Returns:
                bytes
            '''
            return os.read(self.device, length)

        def write(self, data: bytes) -> int:
            '''
            Writes data to the I2C bus

            Args:
                data: the data to write

            Returns:
                The number of bytes written
            '''
            return os.write(self.device, data)

    class DDCInterface(I2CDevice):
        '''
        Class to send DDC (Display Data Channel) commands to an I2C device,
        based on the `Ddcci` and `Mccs` classes from [ddcci.py](https://github.com/siemer/ddcci)
        '''

        PROTOCOL_FLAG = 0x80

        def __init__(self, i2c_path: str):
            '''
            Args:
                i2c_path: the path to the I2C device, eg: `/dev/i2c-2`
            '''
            self.logger = _logger.getChild(self.__class__.__name__).getChild(i2c_path)
            super().__init__(i2c_path, I2C.DDCCI_ADDR)

        def write(self, *args) -> int:
            '''
            Write some data to the I2C device.

            It is recommended to use `setvcp` to set VCP values on the DDC device
            instead of using this function directly.

            Args:
                *args: variable length list of arguments. This will be put
                    into a `bytearray` and wrapped up in various flags and
                    checksums before being written to the I2C device

            Returns:
                The number of bytes that were written
            '''
            time.sleep(I2C.WAIT_TIME)

            ba = bytearray(args)
            ba.insert(0, len(ba) | self.PROTOCOL_FLAG)  # add length info
            ba.insert(0, I2C.HOST_ADDR_W)  # insert source address
            ba.append(functools.reduce(operator.xor, ba, I2C.DESTINATION_ADDR_W))  # checksum

            return super().write(ba)

        def setvcp(self, vcp_code: int, value: int) -> int:
            '''
            Set a VCP value on the device

            Args:
                vcp_code: the VCP command to send, eg: `0x10` is brightness
                value: what to set the value to

            Returns:
                The number of bytes written to the device
            '''
            return self.write(I2C.SET_VCP_CMD, vcp_code, *value.to_bytes(2, 'big'))

        def read(self, amount: int) -> bytes:
            '''
            Reads data from the DDC device.

            It is recommended to use `getvcp` to retrieve VCP values from the
            DDC device instead of using this function directly.

            Args:
                amount: the number of bytes to read

            Raises:
                ValueError: if the read data is deemed invalid
            '''
            time.sleep(I2C.WAIT_TIME)

            ba = super().read(amount + 3)

            # check the bytes read
            checks = {
                'source address': ba[0] == I2C.DESTINATION_ADDR_W,
                'checksum': functools.reduce(operator.xor, ba) == I2C.HOST_ADDR_R,
                'length': len(ba) >= (ba[1] & ~self.PROTOCOL_FLAG) + 3,
            }
            if False in checks.values():
                self.logger.error('i2c read check failed: ' + repr(checks))
                raise I2CValidationError('i2c read check failed: ' + repr(checks))

            return ba[2:-1]

        def getvcp(self, vcp_code: int) -> Tuple[int, int]:
            '''
            Retrieves a VCP value from the DDC device.

            Args:
                vcp_code: the VCP value to read, eg: `0x10` is brightness

            Returns:
                The current and maximum value respectively

            Raises:
                ValueError: if the read data is deemed invalid
            '''
            self.write(I2C.GET_VCP_CMD, vcp_code)
            ba = self.read(8)

            checks = {
                'is feature reply': ba[0] == I2C.GET_VCP_REPLY,
                'supported VCP opcode': ba[1] == 0,
                'answer matches request': ba[2] == vcp_code,
            }
            if False in checks.values():
                self.logger.error('i2c read check failed: ' + repr(checks))
                raise I2CValidationError('i2c read check failed: ' + repr(checks))

            # current and max values
            return int.from_bytes(ba[6:8], 'big'), int.from_bytes(ba[4:6], 'big')

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        all_displays = __cache__.get('i2c_display_info')
        if all_displays is None:
            all_displays = []
            index = 0

            for i2c_path in glob.glob('/dev/i2c-*'):
                if not os.path.exists(i2c_path):
                    continue

                try:
                    # open the I2C device using the host read address
                    device = cls.I2CDevice(i2c_path, cls.HOST_ADDR_R)
                    # read some 512 bytes from the device
                    data = device.read(512)
                except IOError as e:
                    cls._logger.error(f'IOError reading from device {i2c_path}: {e}')
                    continue

                # search for the EDID header within our 512 read bytes
                start = data.find(bytes.fromhex('00 FF FF FF FF FF FF 00'))
                if start < 0:
                    continue

                # grab 128 bytes of the edid
                edid = data[start : start + 128]
                # parse the EDID
                manufacturer_id, manufacturer, model, name, serial = EDID.parse(edid)

                all_displays.append(
                    {
                        'name': name,
                        'model': model,
                        'manufacturer': manufacturer,
                        'manufacturer_id': manufacturer_id,
                        'serial': serial,
                        'method': cls,
                        'index': index,
                        # convert edid to hex string
                        'edid': ''.join(f'{i:02x}' for i in edid),
                        'i2c_bus': i2c_path,
                        'uid': i2c_path.split('-')[-1],
                    }
                )
                index += 1

            if all_displays:
                __cache__.store('i2c_display_info', all_displays, expires=2)

        if display is not None:
            return filter_monitors(display=display, haystack=all_displays, include=['i2c_bus'])
        return all_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        all_displays = cls.get_display_info()
        if display is not None:
            all_displays = [all_displays[display]]

        results = []
        for device in all_displays:
            interface = cls.DDCInterface(device['i2c_bus'])
            value, max_value = interface.getvcp(0x10)

            # make sure display's max brighness is cached
            cache_ident = '%s-%s-%s' % (device['name'], device['model'], device['serial'])
            if cache_ident not in cls._max_brightness_cache:
                cls._max_brightness_cache[cache_ident] = max_value
                cls._logger.info(f'{cache_ident} max brightness:{max_value} (current: {value})')

            if max_value != 100:
                # if max value is not 100 then we have to adjust the scale to be
                # a percentage
                value = int((value / max_value) * 100)

            results.append(value)

        return results

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        all_displays = cls.get_display_info()
        if display is not None:
            all_displays = [all_displays[display]]

        for device in all_displays:
            # make sure display brightness max value is cached
            cache_ident = '%s-%s-%s' % (device['name'], device['model'], device['serial'])
            if cache_ident not in cls._max_brightness_cache:
                cls.get_brightness(display=device['index'])

            # scale the brightness value according to the max brightness
            max_value = cls._max_brightness_cache[cache_ident]
            if max_value != 100:
                value = int((value / 100) * max_value)

            interface = cls.DDCInterface(device['i2c_bus'])
            interface.setvcp(0x10, value)


class XRandr(BrightnessMethodAdv):
    '''
    .. warning:: Deprecated
       The XRandr brightness method has been deprecated and will be removed in a future release.
       XRandr only works on X11, which is starting to be dropped by most distributions.
       Furthermore, the xrandr executable doesn't change backlight brightness like the other methods here,
       but changes the gamma instead, making the behaviour inconsistent with the rest of the library.

    Collection of screen brightness related methods using the xrandr executable
    '''

    executable: str = 'xrandr'
    '''the xrandr executable to be called'''

    @staticmethod
    def _get_uid(interface: str) -> Optional[str]:
        '''
        Attempts to find a UID (I2C bus path) for a given display interface.

        This works by parsing the interface name and matching it up to the entries in `/sys/class/drm`.
        `i2c_bus_from_drm_device` is then used to extract the bus number

        Args:
            interface: the interface in question. EG: `eDP-1`, `eDP1`, `HDMI-1`...

        Returns:
            The bus number as a string if found. Otherwise, none.
        '''
        if not os.path.isdir('/sys/class/drm'):
            return None

        # use regex because sometimes it can be `eDP-1` and sometimes it's `eDP1`
        if interface_match := re.match(r'([a-z]+)-?(\d+)', interface, re.I):
            interface, count = interface_match.groups()
        else:
            return None

        for dir in os.listdir('/sys/class/drm/'):
            # use regex here for case insensitivity on the interface
            if not re.match(r'card\d+-%s(?:-[A-Z])?-%s' % (interface, count), dir, re.I):
                continue
            dir = f'/sys/class/drm/{dir}'
            if bus := i2c_bus_from_drm_device(dir):
                return bus

    @classmethod
    def _gdi(cls):
        '''
        .. warning:: Don't use this
           This function isn't final and I will probably make breaking changes to it.
           You have been warned

        Gets all displays reported by XRandr even if they're not supported
        '''
        xrandr_output = check_output([cls.executable, '--verbose']).decode().split('\n')

        display_count = 0
        tmp_display: dict = {}

        for line_index, line in enumerate(xrandr_output):
            if line == '':
                continue

            if not line.startswith((' ', '\t')) and 'connected' in line and 'disconnected' not in line:
                if tmp_display:
                    yield tmp_display

                tmp_display = {
                    'name': line.split(' ')[0],
                    'interface': line.split(' ')[0],
                    'method': cls,
                    'index': display_count,
                    'model': None,
                    'serial': None,
                    'manufacturer': None,
                    'manufacturer_id': None,
                    'edid': None,
                    'unsupported': line.startswith('XWAYLAND') or 'WAYLAND_DISPLAY' in os.environ,
                    'uid': cls._get_uid(line.split(' ')[0]),
                }
                display_count += 1

            elif 'EDID:' in line:
                # extract the edid from the chunk of the output that will contain the edid
                edid = ''.join(
                    i.replace('\t', '').replace(' ', '') for i in xrandr_output[line_index + 1 : line_index + 9]
                )
                tmp_display['edid'] = edid

                for key, value in zip(
                    ('manufacturer_id', 'manufacturer', 'model', 'name', 'serial'), EDID.parse(tmp_display['edid'])
                ):
                    if value is None:
                        continue
                    tmp_display[key] = value

            elif 'Brightness:' in line:
                tmp_display['brightness'] = int(float(line.replace('Brightness:', '')) * 100)

        if tmp_display:
            yield tmp_display

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None, brightness: bool = False) -> List[dict]:
        '''
        Implements `BrightnessMethod.get_display_info`.

        Args:
            display: the index of the specific display to query.
                If unspecified, all detected displays are queried
            brightness: whether to include the current brightness
                in the returned info
        '''
        valid_displays = []
        for item in cls._gdi():
            if item['unsupported']:
                continue
            if not brightness:
                del item['brightness']
            del item['unsupported']
            valid_displays.append(item)
        if display is not None:
            valid_displays = filter_monitors(display=display, haystack=valid_displays, include=['interface'])
        return valid_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        warnings.warn('The xrandr method is deprecated. Please use another brightness method', DeprecationWarning)
        monitors = cls.get_display_info(brightness=True)
        if display is not None:
            monitors = [monitors[display]]
        brightness = [i['brightness'] for i in monitors]

        return brightness

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        warnings.warn('The xrandr method is deprecated. Please use another brightness method', DeprecationWarning)
        value_as_str = str(float(value) / 100)
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        for i in info:
            check_output([cls.executable, '--output', i['interface'], '--brightness', value_as_str])


class DDCUtil(BrightnessMethodAdv):
    '''collection of screen brightness related methods using the ddcutil executable'''

    _logger = _logger.getChild('DDCUtil')

    executable: str = 'ddcutil'
    '''The ddcutil executable to be called'''
    sleep_multiplier: float = 0.5
    '''
    How long ddcutil should sleep between each DDC request (lower is shorter).
    See [the ddcutil docs](https://www.ddcutil.com/performance_options/#option-sleep-multiplier).
    '''
    cmd_max_tries: int = 10
    '''Max number of retries when calling the ddcutil'''
    enable_async = True
    '''
    Use the `--async` flag when calling ddcutil.
    See [ddcutil docs](https://www.ddcutil.com/performance_options/#option-async)
    '''
    _max_brightness_cache: dict = {}
    '''Cache for displays and their maximum brightness values'''

    @classmethod
    def _gdi(cls):
        '''
        .. warning:: Don't use this
           This function isn't final and I will probably make breaking changes to it.
           You have been warned

        Gets all displays reported by DDCUtil even if they're not supported
        '''
        raw_ddcutil_output = str(
            check_output(
                [cls.executable, 'detect', '-v', f'--sleep-multiplier={cls.sleep_multiplier}'] + ['--async']
                if cls.enable_async
                else [],
                max_tries=cls.cmd_max_tries,
            )
        )[2:-1].split('\\n')
        # Use -v to get EDID string but this means output cannot be decoded.
        # Or maybe it can. I don't know the encoding though, so let's assume it cannot be decoded.
        # Use str()[2:-1] workaround

        # include "Invalid display" sections because they tell us where one displays metadata ends
        # and another begins. We filter out invalid displays later on
        ddcutil_output = [i for i in raw_ddcutil_output if i.startswith(('Invalid display', 'Display', '\t', ' '))]
        tmp_display: dict = {}
        display_count = 0

        for line_index, line in enumerate(ddcutil_output):
            if not line.startswith(('\t', ' ')):
                if tmp_display:
                    yield tmp_display

                tmp_display = {
                    'method': cls,
                    'index': display_count,
                    'model': None,
                    'serial': None,
                    'bin_serial': None,
                    'manufacturer': None,
                    'manufacturer_id': None,
                    'edid': None,
                    'unsupported': 'invalid display' in line.lower(),
                    'uid': None,
                }
                display_count += 1

            elif 'I2C bus' in line:
                tmp_display['i2c_bus'] = line[line.index('/') :]
                tmp_display['bus_number'] = int(tmp_display['i2c_bus'].replace('/dev/i2c-', ''))
                tmp_display['uid'] = tmp_display['i2c_bus'].split('-')[-1]

            elif 'Mfg id' in line:
                # Recently ddcutil has started reporting manufacturer IDs like
                # 'BNQ - UNK' or 'MSI - Microstep' so we have to split the line
                # into chunks of alpha chars and check for a valid mfg id
                for code in re.split(r'[^A-Za-z]', line.replace('Mfg id:', '').replace(' ', '')):
                    if len(code) != 3:
                        # all mfg ids are 3 chars long
                        continue

                    if brand := _monitor_brand_lookup(code):
                        tmp_display['manufacturer_id'], tmp_display['manufacturer'] = brand
                        break

            elif 'Model' in line:
                # the split() removes extra spaces
                name = line.replace('Model:', '').split()
                try:
                    tmp_display['model'] = name[1]
                except IndexError:
                    pass
                tmp_display['name'] = ' '.join(name)

            elif 'Serial number' in line:
                tmp_display['serial'] = line.replace('Serial number:', '').replace(' ', '') or None

            elif 'Binary serial number:' in line:
                tmp_display['bin_serial'] = line.split(' ')[-1][3:-1]

            elif 'EDID hex dump:' in line:
                try:
                    tmp_display['edid'] = ''.join(
                        ''.join(i.split()[1:17]) for i in ddcutil_output[line_index + 2 : line_index + 10]
                    )
                except Exception:
                    pass

        if tmp_display:
            yield tmp_display

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        valid_displays = __cache__.get('ddcutil_monitors_info')
        if valid_displays is None:
            valid_displays = []
            for item in cls._gdi():
                if item['unsupported']:
                    continue
                del item['unsupported']
                valid_displays.append(item)

            if valid_displays:
                __cache__.store('ddcutil_monitors_info', valid_displays)

        if display is not None:
            valid_displays = filter_monitors(display=display, haystack=valid_displays, include=['i2c_bus'])
        return valid_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        monitors = cls.get_display_info()
        if display is not None:
            monitors = [monitors[display]]

        res = []
        for monitor in monitors:
            value = __cache__.get(f'ddcutil_brightness_{monitor["index"]}')
            if value is None:
                cmd_out = (
                    check_output(
                        [
                            cls.executable,
                            'getvcp',
                            '10',
                            '-t',
                            '-b',
                            str(monitor['bus_number']),
                            f'--sleep-multiplier={cls.sleep_multiplier}',
                        ],
                        max_tries=cls.cmd_max_tries,
                    )
                    .decode()
                    .split(' ')
                )

                value = int(cmd_out[-2])
                max_value = int(cmd_out[-1])
                if max_value != 100:
                    # if the max brightness is not 100 then the number is not a percentage
                    # and will need to be scaled
                    value = int((value / max_value) * 100)

                # now make sure max brightness is recorded so set_brightness can use it
                cache_ident = '%s-%s-%s' % (monitor['name'], monitor['serial'], monitor['bin_serial'])
                if cache_ident not in cls._max_brightness_cache:
                    cls._max_brightness_cache[cache_ident] = max_value
                    cls._logger.debug(f'{cache_ident} max brightness:{max_value} (current: {value})')

                __cache__.store(f'ddcutil_brightness_{monitor["index"]}', value, expires=0.5)
            res.append(value)
        return res

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        monitors = cls.get_display_info()
        if display is not None:
            monitors = [monitors[display]]

        __cache__.expire(startswith='ddcutil_brightness_')
        for monitor in monitors:
            # check if monitor has a max brightness that requires us to scale this value
            cache_ident = '%s-%s-%s' % (monitor['name'], monitor['serial'], monitor['bin_serial'])
            if cache_ident not in cls._max_brightness_cache:
                cls.get_brightness(display=monitor['index'])

            if cls._max_brightness_cache[cache_ident] != 100:
                value = int((value / 100) * cls._max_brightness_cache[cache_ident])

            check_output(
                [
                    cls.executable,
                    'setvcp',
                    '10',
                    str(value),
                    '-b',
                    str(monitor['bus_number']),
                    f'--sleep-multiplier={cls.sleep_multiplier}',
                ],
                max_tries=cls.cmd_max_tries,
            )


def i2c_bus_from_drm_device(dir: str) -> Optional[str]:
    '''
    Extract the relevant I2C bus number from a device in `/sys/class/drm`.

    This function works by searching the directory for `i2c-*` and `ddc/i2c-dev/i2c-*` folders.

    Args:
        dir: the DRM directory, in the format `/sys/class/drm/<device>`

    Returns:
        Returns the I2C bus number as a string if found. Otherwise, returns None
    '''
    # check for enabled file and skip device if monitor inactive
    if os.path.isfile(f'{dir}/enabled'):
        with open(f'{dir}/enabled') as f:
            if f.read().strip().lower() != 'enabled':
                return

    # sometimes the i2c path is in /sys/class/drm/*/i2c-*
    # do this first because, in my testing, sometimes a device can have both `.../i2c-X` and `.../ddc/i2c-dev/...`
    # and the latter is usually the wrong i2c path
    paths = glob.glob(f'{dir}/i2c-*')
    if paths:
        return paths[0].split('-')[-1]

    # sometimes the i2c path is in /sys/class/drm/*/ddc/i2c-dev
    if os.path.isdir(f'{dir}/ddc/i2c-dev'):
        paths = os.listdir(f'{dir}/ddc/i2c-dev')
        if paths:
            return paths[0].replace('i2c-', '')


def list_monitors_info(
    method: Optional[str] = None, allow_duplicates: bool = False, unsupported: bool = False
) -> List[dict]:
    '''
    Lists detailed information about all detected displays

    Args:
        method: the method the display can be addressed by. See `.get_methods`
            for more info on available methods
        allow_duplicates: whether to filter out duplicate displays (displays with the same EDID) or not
        unsupported: include detected displays that are invalid or unsupported
    '''
    all_methods = get_methods(method).values()
    haystack = []
    for method_class in all_methods:
        try:
            if unsupported and issubclass(method_class, BrightnessMethodAdv):
                haystack += method_class._gdi()
            else:
                haystack += method_class.get_display_info()
        except Exception as e:
            _logger.warning(f'error grabbing display info from {method_class} - {format_exc(e)}')
            pass

    if allow_duplicates:
        return haystack

    try:
        # use filter_monitors to remove duplicates
        return filter_monitors(haystack=haystack)
    except NoValidDisplayError:
        return []


METHODS = (SysFiles, I2C, XRandr, DDCUtil)



'''
Contains globally applicable configuration variables.
'''

from functools import wraps
from typing import Callable, Optional


def default_params(func: Callable):
    '''
    This decorator sets default kwarg values using global configuration variables.
    '''

    @wraps(func)
    def wrapper(*args, **kwargs):
        kwargs.setdefault('allow_duplicates', ALLOW_DUPLICATES)
        kwargs.setdefault('method', METHOD)
        return func(*args, **kwargs)

    return wrapper


ALLOW_DUPLICATES: bool = False
'''
Default value for the `allow_duplicates` parameter in top-level functions.
'''

METHOD: Optional[str] = None
'''
Default value for the `method` parameter in top-level functions.

For available values, see `.get_methods`
'''



import logging
import platform
import threading
import time
import traceback
from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Dict, FrozenSet, List, Optional, Tuple, Type, Union

from . import config
from ._version import __author__, __version__  # noqa: F401
from .exceptions import NoValidDisplayError, format_exc
from .helpers import BrightnessMethod, ScreenBrightnessError, logarithmic_range, percentage
from .types import DisplayIdentifier, IntPercentage, Percentage

_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())


@config.default_params
def get_brightness(
    display: Optional[DisplayIdentifier] = None,
    method: Optional[str] = None,
    allow_duplicates: Optional[bool] = None,
    verbose_error: bool = False,
) -> List[Union[IntPercentage, None]]:
    '''
    Returns the current brightness of one or more displays

    Args:
        display (.types.DisplayIdentifier): the specific display to query
        method: the method to use to get the brightness. See `get_methods` for
            more info on available methods
        allow_duplicates: controls whether to filter out duplicate displays or not.
        verbose_error: controls the level of detail in the error messages

    Returns:
        A list of `.types.IntPercentage` values, each being the brightness of an
        individual display. Invalid displays may return None.

    Example:
        ```python
        import screen_brightness_control as sbc

        # get the current screen brightness (for all detected displays)
        current_brightness = sbc.get_brightness()

        # get the brightness of the primary display
        primary_brightness = sbc.get_brightness(display=0)

        # get the brightness of the secondary display (if connected)
        secondary_brightness = sbc.get_brightness(display=1)
        ```
    '''
    result = __brightness(
        display=display,
        method=method,
        meta_method='get',
        allow_duplicates=allow_duplicates,
        verbose_error=verbose_error,
    )
    # __brightness can return None depending on the `no_return` kwarg. That obviously would never happen here
    # but the type checker doesn't see it that way.
    return [] if result is None else result


@config.default_params
def set_brightness(
    value: Percentage,
    display: Optional[DisplayIdentifier] = None,
    method: Optional[str] = None,
    force: bool = False,
    allow_duplicates: Optional[bool] = None,
    verbose_error: bool = False,
    no_return: bool = True,
) -> Optional[List[Union[IntPercentage, None]]]:
    '''
    Sets the brightness level of one or more displays to a given value.

    Args:
        value (.types.Percentage): the new brightness level
        display (.types.DisplayIdentifier): the specific display to adjust
        method: the method to use to set the brightness. See `get_methods` for
            more info on available methods
        force: [*Linux Only*] if False the brightness will never be set lower than 1.
            This is because on most displays a brightness of 0 will turn off the backlight.
            If True, this check is bypassed
        allow_duplicates: controls whether to filter out duplicate displays or not.
        verbose_error: boolean value controls the amount of detail error messages will contain
        no_return: don't return the new brightness level(s)

    Returns:
        If `no_return` is set to `True` (the default) then this function returns nothing.
        Otherwise, a list of `.types.IntPercentage` is returned, each item being the new
        brightness of each adjusted display (invalid displays may return None)

    Example:
        ```python
        import screen_brightness_control as sbc

        # set brightness to 50%
        sbc.set_brightness(50)

        # set brightness to 0%
        sbc.set_brightness(0, force=True)

        # increase brightness by 25%
        sbc.set_brightness('+25')

        # decrease brightness by 30%
        sbc.set_brightness('-30')

        # set the brightness of display 0 to 50%
        sbc.set_brightness(50, display=0)
        ```
    '''
    if isinstance(value, str) and ('+' in value or '-' in value):
        output: List[Union[IntPercentage, None]] = []
        for monitor in filter_monitors(display=display, method=method, allow_duplicates=allow_duplicates):
            # `filter_monitors()` will raise an error if no valid displays are found
            display_instance = Display.from_dict(monitor)
            display_instance.set_brightness(value=value, force=force)
            output.append(None if no_return else display_instance.get_brightness())

        return None if no_return else output

    if platform.system() == 'Linux' and not force:
        lower_bound = 1
    else:
        lower_bound = 0

    value = percentage(value, lower_bound=lower_bound)

    return __brightness(
        value,
        display=display,
        method=method,
        meta_method='set',
        no_return=no_return,
        allow_duplicates=allow_duplicates,
        verbose_error=verbose_error,
    )


@config.default_params
def fade_brightness(
    finish: Percentage,
    start: Optional[Percentage] = None,
    interval: float = 0.01,
    increment: int = 1,
    blocking: bool = True,
    force: bool = False,
    logarithmic: bool = True,
    stoppable: bool = True,
    **kwargs,
) -> Union[List[threading.Thread], List[Union[IntPercentage, None]]]:
    '''
    Gradually change the brightness of one or more displays

    Args:
        finish (.types.Percentage): fade to this brightness level
        start (.types.Percentage): where the brightness should fade from.
            If this arg is not specified, the fade will be started from the
            current brightness.
        interval: the time delay between each step in brightness
        increment: the amount to change the brightness by per step
        blocking: whether this should occur in the main thread (`True`) or a new daemonic thread (`False`)
        force: [*Linux Only*] if False the brightness will never be set lower than 1.
            This is because on most displays a brightness of 0 will turn off the backlight.
            If True, this check is bypassed
        logarithmic: follow a logarithmic brightness curve when adjusting the brightness
        stoppable: whether the fade can be stopped by starting a new fade on the same display
        **kwargs: passed through to `filter_monitors` for display selection.
            Will also be passed to `get_brightness` if `blocking is True`

    Returns:
        By default, this function calls `get_brightness()` to return the new
        brightness of any adjusted displays.

        If `blocking` is set to `False`, then a list of threads are
        returned, one for each display being faded.

    Example:
        ```python
        import screen_brightness_control as sbc

        # fade brightness from the current brightness to 50%
        sbc.fade_brightness(50)

        # fade the brightness from 25% to 75%
        sbc.fade_brightness(75, start=25)

        # fade the brightness from the current value to 100% in steps of 10%
        sbc.fade_brightness(100, increment=10)

        # fade the brightness from 100% to 90% with time intervals of 0.1 seconds
        sbc.fade_brightness(90, start=100, interval=0.1)

        # fade the brightness to 100% in a new thread
        sbc.fade_brightness(100, blocking=False)
        ```
    '''
    # make sure only compatible kwargs are passed to filter_monitors
    available_monitors = filter_monitors(
        **{k: v for k, v in kwargs.items() if k in ('display', 'haystack', 'method', 'include', 'allow_duplicates')}
    )

    threads = []
    for i in available_monitors:
        display = Display.from_dict(i)

        thread = threading.Thread(
            target=display._fade_brightness,
            args=(finish,),
            kwargs={
                'start': start,
                'interval': interval,
                'increment': increment,
                'force': force,
                'logarithmic': logarithmic,
                'stoppable': stoppable,
            },
        )
        thread.start()
        threads.append(thread)

    if not blocking:
        return threads

    for t in threads:
        t.join()
    return get_brightness(**kwargs)


@config.default_params
def list_monitors_info(
    method: Optional[str] = None, allow_duplicates: Optional[bool] = None, unsupported: bool = False
) -> List[dict]:
    '''
    List detailed information about all displays that are controllable by this library

    Args:
        method: the method to use to list the available displays. See `get_methods` for
            more info on available methods
        allow_duplicates: controls whether to filter out duplicate displays or not.
        unsupported: include detected displays that are invalid or unsupported

    Returns:
        list: list of dictionaries containing information about the detected displays

    Example:
        ```python
        import screen_brightness_control as sbc
        displays = sbc.list_monitors_info()
        for display in displays:
            print('=======================')
            # the manufacturer name plus the model
            print('Name:', display['name'])
            # the general model of the display
            print('Model:', display['model'])
            # the serial of the display
            print('Serial:', display['serial'])
            # the name of the brand of the display
            print('Manufacturer:', display['manufacturer'])
            # the 3 letter code corresponding to the brand name, EG: BNQ -> BenQ
            print('Manufacturer ID:', display['manufacturer_id'])
            # the index of that display FOR THE SPECIFIC METHOD THE DISPLAY USES
            print('Index:', display['index'])
            # the method this display can be addressed by
            print('Method:', display['method'])
            # the EDID string associated with that display
            print('EDID:', display['edid'])
            # The UID of the display
            print('UID:', display['uid'])
        ```
    '''
    return _OS_MODULE.list_monitors_info(method=method, allow_duplicates=allow_duplicates, unsupported=unsupported)


@config.default_params
def list_monitors(method: Optional[str] = None, allow_duplicates: Optional[bool] = None) -> List[str]:
    '''
    List the names of all detected displays

    Args:
        method: the method to use to list the available displays. See `get_methods` for
            more info on available methods
        allow_duplicates: controls whether to filter out duplicate displays or not.

    Example:
        ```python
        import screen_brightness_control as sbc
        display_names = sbc.list_monitors()
        # eg: ['BenQ GL2450H', 'Dell U2211H']
        ```
    '''
    return [i['name'] for i in list_monitors_info(method=method, allow_duplicates=allow_duplicates)]


def get_methods(name: Optional[str] = None) -> Dict[str, Type[BrightnessMethod]]:
    '''
    Returns all available brightness method names and their associated classes.

    Args:
        name: if specified, return the method corresponding to this name

    Raises:
        ValueError: if the given name is incorrect

    Example:
        ```python
        import screen_brightness_control as sbc

        all_methods = sbc.get_methods()

        for method_name, method_class in all_methods.items():
            print('Method:', method_name)
            print('Class:', method_class)
            print('Associated monitors:', sbc.list_monitors(method=method_name))
        ```
    '''
    methods: Dict[str, type[BrightnessMethod]] = {i.__name__.lower(): i for i in _OS_MODULE.METHODS}

    if name is None:
        return methods

    if not isinstance(name, str):
        raise TypeError(f'name must be of type str, not {type(name)!r}')

    name = name.lower()
    if name in methods:
        return {name: methods[name]}

    _logger.debug(f'requested method {name!r} invalid')
    raise ValueError(f'invalid method {name!r}, must be one of: {list(methods)}')


@dataclass
class Display:
    '''
    Represents a single connected display.
    '''

    index: int
    '''The index of the display relative to the method it uses.
    So if the index is 0 and the method is `windows.VCP`, then this is the 1st
    display reported by `windows.VCP`, not the first display overall.'''
    method: Type[BrightnessMethod]
    '''The method by which this monitor can be addressed.
    This will be a class from either the windows or linux sub-module'''

    uid: Optional[str] = None
    '''A unique identifier for the display. This is usually inferred from the display's connection to the machine.'''
    edid: Optional[str] = None
    '''A 256 character hex string containing information about a display and its capabilities'''
    manufacturer: Optional[str] = None
    '''Name of the display's manufacturer'''
    manufacturer_id: Optional[str] = None
    '''3 letter code corresponding to the manufacturer name'''
    model: Optional[str] = None
    '''Model name of the display'''
    name: Optional[str] = None
    '''The name of the display, often the manufacturer name plus the model name'''
    serial: Optional[str] = None
    '''The serial number of the display or (if serial is not available) an ID assigned by the OS'''

    _logger: logging.Logger = field(init=False, repr=False)
    _fade_thread_dict: ClassVar[Dict[FrozenSet[Any], threading.Thread]] = {}
    '''A dictionary mapping display identifiers to latest fade threads for stopping fades.'''

    def __post_init__(self):
        self._logger = _logger.getChild(self.__class__.__name__).getChild(str(self.get_identifier()[1])[:20])

    def fade_brightness(
        self,
        finish: Percentage,
        start: Optional[Percentage] = None,
        interval: float = 0.01,
        increment: int = 1,
        force: bool = False,
        logarithmic: bool = True,
        blocking: bool = True,
        stoppable: bool = True,
    ) -> Optional[threading.Thread]:
        '''
        Gradually change the brightness of this display to a set value.
        Can execute in the current thread, blocking until completion,
        or in a separate thread, allowing concurrent operations.
        When set as non-blocking and stoppable, a new fade can halt the this operation.

        Args:
            finish (.types.Percentage): the brightness level to end up on
            start (.types.Percentage): where the fade should start from. Defaults
                to whatever the current brightness level for the display is
            interval: time delay between each change in brightness
            increment: amount to change the brightness by each time (as a percentage)
            force: [*Linux only*] allow the brightness to be set to 0. By default,
                brightness values will never be set lower than 1, since setting them to 0
                often turns off the backlight
            logarithmic: follow a logarithmic curve when setting brightness values.
                See `logarithmic_range` for rationale
            blocking: run this function in the current thread and block until it completes
            stoppable: whether this fade will be stopped by starting a new fade on the same display

        Returns:
            If `blocking` is `False`, returns a `threading.Thread` object representing the
            thread in which the fade operation is running.
            Otherwise, it returns None.
        '''
        thread = threading.Thread(
            target=self._fade_brightness,
            args=(finish,),
            kwargs={
                'start': start,
                'interval': interval,
                'increment': increment,
                'force': force,
                'logarithmic': logarithmic,
                'stoppable': stoppable,
            },
        )
        thread.start()

        if not blocking:
            return thread
        else:
            thread.join()
            return None

    def _fade_brightness(
        self,
        finish: Percentage,
        start: Optional[Percentage] = None,
        interval: float = 0.01,
        increment: int = 1,
        force: bool = False,
        logarithmic: bool = True,
        stoppable: bool = True,
    ) -> None:
        # Record the latest thread for this display so that other stoppable threads can be stopped
        display_key = frozenset((self.method, self.index))
        self._fade_thread_dict[display_key] = threading.current_thread()
        # minimum brightness value
        if platform.system() == 'Linux' and not force:
            lower_bound = 1
        else:
            lower_bound = 0

        current = self.get_brightness()

        finish = percentage(finish, current, lower_bound)
        start = percentage(current if start is None else start, current, lower_bound)

        # mypy says "object is not callable" but range is. Ignore this
        range_func: Callable = logarithmic_range if logarithmic else range  # type: ignore[assignment]
        increment = abs(increment)
        if start > finish:
            increment = -increment

        self._logger.debug(f'fade {start}->{finish}:{increment}:logarithmic={logarithmic}')

        # Record the time when the next brightness change should start
        next_change_start_time = time.time()
        for value in range_func(start, finish, increment):
            if stoppable and threading.current_thread() != self._fade_thread_dict[display_key]:
                # If the current thread is stoppable and it's not the latest thread, stop fading
                break
            # `value` is ensured not to hit `finish` in loop, this will be handled in the final step.
            self.set_brightness(value, force=force)

            # `interval` is the intended time between the start of each brightness change.
            next_change_start_time += interval
            sleep_time = next_change_start_time - time.time()
            # Skip sleep if the scheduled time has already passed
            if sleep_time > 0:
                time.sleep(sleep_time)
        else:
            # As `value` doesn't hit `finish` in loop, we explicitly set brightness to `finish`.
            # This also avoids an unnecessary sleep in the last iteration.
            if not stoppable or threading.current_thread() == self._fade_thread_dict[display_key]:
                self.set_brightness(finish, force=force)

    @classmethod
    def from_dict(cls, display: dict) -> 'Display':
        '''
        Initialise an instance of the class from a dictionary, ignoring
        any unwanted keys
        '''
        return cls(
            index=display['index'],
            method=display['method'],
            edid=display['edid'],
            manufacturer=display['manufacturer'],
            manufacturer_id=display['manufacturer_id'],
            model=display['model'],
            name=display['name'],
            serial=display['serial'],
            uid=display.get('uid'),
        )

    def get_brightness(self) -> IntPercentage:
        '''
        Returns the brightness of this display.

        Returns:
            The brightness value of the display, as a percentage.
            See `.types.IntPercentage`
        '''
        return self.method.get_brightness(display=self.index)[0]

    def get_identifier(self) -> Tuple[str, DisplayIdentifier]:
        '''
        Returns the `.types.DisplayIdentifier` for this display.
        Will iterate through the UID, EDID, serial, name and index and return the first
        value that is not equal to None

        Returns:
            The name of the property returned and the value of said property.
            EG: `('serial', '123abc...')` or `('name', 'BenQ GL2450H')`
        '''
        for key in ('uid', 'edid', 'serial', 'name'):
            value = getattr(self, key, None)
            if value is not None:
                return key, value
        # the index should surely never be `None`
        return 'index', self.index

    def is_active(self) -> bool:
        '''
        Attempts to retrieve the brightness for this display. If it works the display is deemed active
        '''
        try:
            self.get_brightness()
            return True
        except Exception as e:
            self._logger.error(
                f'Display.is_active: {self.get_identifier()} failed get_brightness call - {format_exc(e)}'
            )
            return False

    def set_brightness(self, value: Percentage, force: bool = False):
        '''
        Sets the brightness for this display. See `set_brightness` for the full docs

        Args:
            value (.types.Percentage): the brightness percentage to set the display to
            force: allow the brightness to be set to 0 on Linux. This is disabled by default
                because setting the brightness of 0 will often turn off the backlight
        '''
        # convert brightness value to percentage
        if platform.system() == 'Linux' and not force:
            lower_bound = 1
        else:
            lower_bound = 0

        value = percentage(value, current=self.get_brightness, lower_bound=lower_bound)

        self.method.set_brightness(value, display=self.index)


@config.default_params
def filter_monitors(
    display: Optional[DisplayIdentifier] = None,
    haystack: Optional[List[dict]] = None,
    method: Optional[str] = None,
    include: List[str] = [],
    allow_duplicates: Optional[bool] = None,
) -> List[dict]:
    '''
    Searches through the information for all detected displays
    and attempts to return the info matching the value given.
    Will attempt to match against index, name, edid, method and serial

    Args:
        display (.types.DisplayIdentifier): the display you are searching for
        haystack: the information to filter from.
            If this isn't set it defaults to the return of `list_monitors_info`
        method: the method the monitors use. See `get_methods` for
            more info on available methods
        include: extra fields of information to sort by
        allow_duplicates: controls whether to filter out duplicate displays or not

    Raises:
        NoValidDisplayError: if the display does not have a match
        TypeError: if the `display` kwarg is not `int` or `str`

    Example:
        ```python
        import screen_brightness_control as sbc

        search = 'GL2450H'
        match = sbc.filter_displays(search)
        print(match)
        # EG output: [{'name': 'BenQ GL2450H', 'model': 'GL2450H', ... }]
        ```
    '''
    if display is not None and type(display) not in (str, int):
        raise TypeError(f'display kwarg must be int or str, not \"{type(display).__name__}\"')

    def get_monitor_list():
        # if we have been provided with a list of monitors to sift through then use that
        # otherwise, get the info ourselves
        if haystack is not None:
            monitors_with_duplicates = haystack
            if method is not None:
                method_class = next(iter(get_methods(method).values()))
                monitors_with_duplicates = [i for i in haystack if i['method'] == method_class]
        else:
            monitors_with_duplicates = list_monitors_info(method=method, allow_duplicates=True)

        return monitors_with_duplicates

    def filter_monitor_list(to_filter):
        # This loop does two things:
        # 1. Filters out duplicate monitors
        # 2. Matches the display kwarg (if applicable)

        # When duplicates are allowed, the logic is straightforward:
        if allow_duplicates:
            if display is None:
                # no monitor should be filtered out
                return to_filter
            elif isinstance(display, int):
                # 'display' variable should be the index of the monitor
                # return a list with the monitor at the index or an empty list if the index is out of range
                return to_filter[display : display + 1]
            elif isinstance(display, str):
                # 'display' variable should be an identifier of the monitor
                # multiple monitors with the same identifier are allowed here, so return all of them
                monitors = []
                for monitor in to_filter:
                    for identifier in ['uid', 'edid', 'serial', 'name'] + include:
                        if display == monitor.get(identifier, None):
                            monitors.append(monitor)
                            break
                return monitors

        filtered_displays = {}
        for monitor in to_filter:
            # find a valid identifier for a monitor, excluding any which are equal to None
            added = False
            for identifier in ['uid', 'edid', 'serial', 'name'] + include:
                if monitor.get(identifier, None) is None:
                    continue

                m_id = monitor[identifier]
                if m_id in filtered_displays:
                    break

                if isinstance(display, str) and m_id != display:
                    continue

                # check we haven't already added the monitor
                if not added:
                    filtered_displays[m_id] = monitor
                    added = True

                # if the display kwarg is an integer and we are currently at that index
                if isinstance(display, int) and len(filtered_displays) - 1 == display:
                    return [monitor]

                if added:
                    break
        return list(filtered_displays.values())

    duplicates = []
    for _ in range(3):
        duplicates = get_monitor_list()
        if duplicates:
            break
        time.sleep(0.4)
    else:
        msg = 'no displays detected'
        if method is not None:
            msg += f' with method: {method!r}'
        raise NoValidDisplayError(msg)

    monitors = filter_monitor_list(duplicates)
    if not monitors:
        # if no displays matched the query
        msg = 'no displays found'
        if display is not None:
            msg += f' with name/serial/edid/index of {display!r}'
        if method is not None:
            msg += f' with method of {method!r}'
        raise NoValidDisplayError(msg)

    return monitors


def __brightness(
    *args: Any,
    display: Optional[DisplayIdentifier] = None,
    method: Optional[str] = None,
    meta_method: str = 'get',
    no_return: bool = False,
    allow_duplicates: bool = False,
    verbose_error: bool = False,
    **kwargs: Any,
) -> Optional[List[Union[IntPercentage, None]]]:
    '''Internal function used to get/set brightness'''
    _logger.debug(f'brightness {meta_method} request display {display} with method {method}')

    output: List[Union[int, None]] = []
    errors = []

    for monitor in filter_monitors(display=display, method=method, allow_duplicates=allow_duplicates):
        try:
            if meta_method == 'set':
                monitor['method'].set_brightness(*args, display=monitor['index'], **kwargs)
                if no_return:
                    output.append(None)
                    continue

            output += monitor['method'].get_brightness(display=monitor['index'], **kwargs)
        except Exception as e:
            output.append(None)
            errors.append((monitor, e.__class__.__name__, traceback.format_exc() if verbose_error else e))

    if output:
        output_is_none = set(output) == {None}
        if (
            # can't have None output if we are trying to get the brightness
            (meta_method == 'get' and not output_is_none)
            or (
                # if we are setting the brightness then we CAN have a None output
                # but only if no_return is True.
                meta_method == 'set' and ((no_return and output_is_none) or not output_is_none)
            )
        ):
            return None if no_return else output

    # if the function hasn't returned then it has failed
    msg = '\n'
    if errors:
        for monitor, exc_name, exc in errors:
            if isinstance(monitor, str):
                msg += f'\t{monitor}'
            else:
                msg += f'\t{monitor["name"]} ({monitor["serial"]})'
            msg += f' -> {exc_name}: '
            msg += str(exc).replace('\n', '\n\t\t') + '\n'
    else:
        msg += '\tno valid output was received from brightness methods'

    raise ScreenBrightnessError(msg)


if platform.system() == 'Windows':
    from . import windows

    _OS_MODULE = windows
elif platform.system() == 'Linux':
    from . import linux

    _OS_MODULE = linux
else:
    _logger.warning(f'package imported on unsupported platform ({platform.system()})')



import subprocess


def format_exc(e: Exception) -> str:
    '''@private'''
    return f'{type(e).__name__}: {e}'


class ScreenBrightnessError(Exception):
    '''
    Generic error class designed to make catching errors under one umbrella easy.
    '''

    ...


class EDIDParseError(ScreenBrightnessError):
    '''Unparsable/invalid EDID'''

    ...


class NoValidDisplayError(ScreenBrightnessError, LookupError):
    '''Could not find a valid display'''

    ...


class I2CValidationError(ScreenBrightnessError):
    '''I2C data validation failed'''

    ...


class MaxRetriesExceededError(ScreenBrightnessError, subprocess.CalledProcessError):
    '''
    The command has been retried too many times.

    Example:
        ```python
        try:
            subprocess.check_output(['exit', '1'])
        except subprocess.CalledProcessError as e:
            raise MaxRetriesExceededError('failed after 1 try', e)
        ```
    '''

    def __init__(self, message: str, exc: subprocess.CalledProcessError):
        self.message: str = message
        ScreenBrightnessError.__init__(self, message)
        super().__init__(exc.returncode, exc.cmd, exc.stdout, exc.stderr)

    def __str__(self):
        string = super().__str__()
        string += f'\n\t-> {self.message}'
        return string
